import numpy as np
import cvxpy as cp
import baryrat
from scipy.optimize import minimize


def cvx_matrix_projection(
    zM, GM_matrix, w_cut=10, n_real=201, ofile='Giw', solver='SCS', **kwargs
):
    """Projection of noisy GW data on to Nevanlinna manifold.
    Once this is performed, analytic continuation either using Nevanlinna
    or ES becomes easier.
    Input arguments:
        zM              :   n_imag number of imaginary frequency values
        GM_matrix       :   Exact Matsubara Green's function data in the shape
                            (n_imag, n_orb, n_orb)
        w_cut           :   cut-off to consider for real axis
        n_real          :   number of real frequencies to consider
    Returns:
        G_iw_proj       :   Projected Green's function
    such that
        G(iw_m) = sum_n P_n / (iw_m - w_n)
    """
    # number of imag frequency points
    n_imag = len(zM)

    # number of orbitals
    n_orb = GM_matrix.shape[-1]
    assert GM_matrix.shape == (n_imag, n_orb, n_orb)

    # real frequency grid
    w_grid = np.linspace(-w_cut, w_cut, n_real)

    # Initialize the cvxpy variables P_vec
    # These variables are positive-semidefinite (PSD)
    P_vec = cp.Variable((n_real, n_orb * n_orb), complex=True)

    # PSD and Hermitian constraint
    constr = []
    for i in range(n_real):
        constr.append(
            cp.reshape(P_vec[i, :], (n_orb, n_orb)) >> 0
        )

    #
    # Define the objective function
    #

    # real and imaginary mesh matrix
    zw_matrix = np.zeros((n_imag, n_real), dtype=complex)
    for i, iw in enumerate(zM):
        for j, w in enumerate(w_grid):
            zw_matrix[i, j] = 1 / (iw - w)

    G_approx = zw_matrix @ P_vec
    G_diff = G_approx - GM_matrix.reshape((n_imag, n_orb * n_orb))
    obj_qty = cp.norm(G_diff, p='fro')

    #
    # Define cvxy's Objective problem and solve it
    #

    prob = cp.Problem(cp.Minimize(obj_qty), constr)
    # NOTE: using default convergence criterion for the solver
    opt_error = prob.solve(solver=solver, **kwargs)
    print("Error CVXPy optimization: ", opt_error)
    print("Objective value after optimization: ", obj_qty.value)

    #
    # Get results for the projected imaginary-time Greens function
    #

    np_P_vec = P_vec.value
    G_iw_proj = zw_matrix @ np_P_vec
    np.savetxt(ofile, G_iw_proj)

    return


def cvx_diag_projection(
    zM, GM_diag, w_cut=10, n_real=201, ofile='Giw', solver='SCS', **kwargs
):
    """Projection of noisy GW data on to Nevanlinna manifold.
    Once this is performed, analytic continuation either using Nevanlinna
    or ES becomes easier.
    Input arguments:
        zM              :   n_imag number of imaginary frequency values
        GM_diag         :   Exact Matsubara Green's function data in the shape
                            (n_imag, n_orb)
        w_cut           :   cut-off to consider for real axis
        n_real          :   number of real frequencies to consider
    Returns:
        G_iw_proj       :   Projected Green's function
    such that
        G(iw_m) = sum_n P_n / (iw_m - w_n)
    """
    # number of imag frequency points
    n_imag = len(zM)

    # number of orbitals
    n_orb = GM_diag.shape[-1]
    assert GM_diag.shape == (n_imag, n_orb)

    # real frequency grid
    w_grid = np.linspace(-w_cut, w_cut, n_real)

    # Initialize the cvxpy variables P_vec
    # These variables are non-negative vectors
    P_vec = cp.Variable((n_real, n_orb), nonneg=True)

    #
    # Define the objective function
    #

    # real and imaginary mesh matrix
    zw_matrix = np.zeros((n_imag, n_real), dtype=complex)
    for i, iw in enumerate(zM):
        for j, w in enumerate(w_grid):
            zw_matrix[i, j] = 1 / (iw - w)

    G_approx = zw_matrix @ P_vec
    G_diff = G_approx - GM_diag
    obj_qty = cp.norm(G_diff, p='fro')

    #
    # Define cvxy's Objective problem and solve it
    #

    prob = cp.Problem(cp.Minimize(obj_qty))
    # NOTE: using default convergence criterion for the solver
    opt_error = prob.solve(solver=solver, **kwargs)
    print("Error CVXPy optimization: ", opt_error)
    print("Objective value after optimization: ", obj_qty.value)

    #
    # Get results for the projected imaginary-time Greens function
    #

    np_P_vec = P_vec.value
    G_iw_proj = zw_matrix @ np_P_vec
    G_iw_proj = G_iw_proj.reshape((n_imag, n_orb))
    np.savetxt(ofile, G_iw_proj)

    return


def cvx_optimize(poles, GM_matrix, zM, solver='SCS', **kwargs):
    """Top level target error function in the ES approach
    to Nevanlinna, i.e.,
        Err (poles) = min_{X} || Gimag (iw) - Gapprox [poles, X] (iw) ||
    Input Arguments:
        poles           :   pole positions
        GM_matrix       :   Exact Matsubara Green's function data in the shape
                            (num_imag, num_orb, num_orb)
        zM              :   Matsubara frequencies
        solver          :   Specify solver type for ES SDR fit
    Returns:
        opt_error       :   Optimal value of Err (poles) for fixed poles
        np_X_vec        :   Numpy array rep of optimal X_vec in the shape
                            (num_poles, num_orb, num_orb)
        np_G_approxx    :   Numpy array rep of approximated optimal Green's
                            function for the specified Matsubara frequencies.
                            The shape of the array is
                            (num_imag, num_orb, num_orb)
    """
    # number of imag frequency points
    num_imag = len(zM)
    # Get the number of poles.
    num_poles = len(poles)
    # number of orbitals
    num_orb = GM_matrix.shape[-1]

    # Initialize the cvxpy variables X_vec.
    # These variables are positive-semidefinite (PSD) and Hermitian matrices.
    # Loosely speaking, the X variables are of the shape
    #   X_vec = np.zeros((Num_poles, N_orb, N_orb), dtype=complex)
    X_vec = cp.Variable((num_poles, num_orb * num_orb), complex=False)

    # Operator for transposse of X_vec matrices
    # We will use this operator to enforce Hermitian nature.
    X_trans = np.zeros((num_orb * num_orb, num_orb * num_orb))
    for i in range(num_orb * num_orb):
        j = num_orb * (i % num_orb) + i // num_orb
        X_trans[j, i] = 1.0

    # PSD and Hermitian constraint
    constr = []
    for i in range(num_poles):
        constr.append(
            cp.reshape(X_vec[i, :], (num_orb, num_orb)) >> 0
        )

    #
    # Define the objective function
    #

    # construct iw - poles matrix
    iw_pole_matrix = np.zeros((num_imag, num_poles), dtype=complex)
    for iw in range(num_imag):
        for m in range(num_poles):
            iw_pole_matrix[iw, m] = 1 / (1j * zM[iw] - poles[m])

    # Construct G_approx of shape ((N_orb, N_orb, Num_poles), dtype=complex)
    G_approx = iw_pole_matrix @ X_vec \
        + iw_pole_matrix @ cp.conj(X_vec) @ X_trans
    obj_qty = cp.norm(
        G_approx - GM_matrix.reshape((num_imag, num_orb * num_orb)),
        p='fro'
    )

    #
    # Define cvxy's Objective problem and solve it
    #

    prob = cp.Problem(cp.Minimize(obj_qty), constr)
    # NOTE: using default convergence criterion for the solver
    opt_error = prob.solve(solver=solver, **kwargs)
    # print("Constraint values: ")
    # constr_values = [constr[i].dual_value for i in range(num_poles)]
    # print(constr_values)

    #
    # Get results and store in np_X_vec
    #

    # X vectors
    np_X_vec = X_vec.value
    np_X_vec += np.conj(np_X_vec) @ X_trans
    np_X_vec = np_X_vec.reshape((num_poles, num_orb, num_orb))
    # approximated Green's function on iw points
    np_G_approx = np.einsum('wp, pij -> wij', iw_pole_matrix, np_X_vec)

    return opt_error, np_X_vec, np_G_approx


def cvx_optimize_spectral(poles, GM_diags, zM, solver='SCS', **kwargs):
    """Top level target error function in the ES approach
    to Nevanlinna, i.e.,
        Err (poles) = min_{X} || Gimag (iw) - Gapprox [poles, X] (iw) ||
    This function only computes the spectral function, not the full G matrix.
    Input Arguments:
        poles           :   pole positions
        GM_diags        :   Exact Matsubara Green's function diagonal data
                            in the shape (num_imag, num_orb)
        zM              :   Matsubara frequencies
        solver          :   Specify solver for ES SDR fit
    Returns:
        opt_error       :   Optimal value of Err (poles) for fixed poles
        np_X_vec        :   Numpy array rep of optimal X_vec in the shape
                            (num_poles, num_orb)
        np_G_approxx    :   Numpy array rep of approximated optimal Green's
                            function for the specified Matsubara frequencies.
                            The shape of the array is
                            (num_imag, num_orb)
    """
    # number of imag frequency points
    num_imag = len(zM)
    # Get the number of poles.
    num_poles = len(poles)
    # number of orbitals
    num_orb = GM_diags.shape[-1]

    # Initialize the cvxpy variables X_vec.
    # These variables are positive-semidefinite (PSD) and Hermitian matrices.
    # Loosely speaking, the X variables are of the shape
    #   X_vec = np.zeros((Num_poles, N_orb, N_orb), dtype=complex)
    X_vec = cp.Variable((num_poles, num_orb), nonneg=True)

    #
    # Define the objective function
    #

    # construct iw - poles matrix
    iw_pole_matrix = np.zeros((num_imag, num_poles), dtype=complex)
    for iw in range(num_imag):
        for m in range(num_poles):
            iw_pole_matrix[iw, m] = 1 / (1j * zM[iw] - poles[m])

    # Construct G_approx of shape ((N_orb, N_orb, Num_poles), dtype=complex)
    G_approx = iw_pole_matrix @ X_vec
    obj_qty = cp.norm(
        G_approx - GM_diags.reshape((num_imag, num_orb)),
        p='fro'
    )

    #
    # Define cvxy's Objective problem and solve itjj
    #

    prob = cp.Problem(cp.Minimize(obj_qty))
    # NOTE: using default convergence criterion for the solver
    opt_error = prob.solve(solver=solver, **kwargs)
    # print("Constraint values: ")
    # constr_values = [constr[i].dual_value for i in range(num_poles)]
    # print(constr_values)

    #
    # Get results and store in np_X_vec
    #

    # X vectors
    np_X_vec = X_vec.value
    np_X_vec = np_X_vec.reshape((num_poles, num_orb))
    # approximated Green's function on iw points
    np_G_approx = iw_pole_matrix @ np_X_vec

    return opt_error, np_X_vec, np_G_approx


def cvx_gradient(poles, GM_matrix, zM):
    """Computes the gradient for top-level optimization of pole positions
    in ES approach to Nevanlinna
        d Err / d poles
    where,
        Err (poles) = min_{X} || Gimag (iw) - Gapprox [poles, X] (iw) ||
    Specifically, G_approx is defined as
        G_approx = sum_m X_vec(:,:,m)/(1j*zM - poles(m))
        X_vec is shape (N_orb, N_orb, num_poles) and is just an array of
        optimized X_l matrices, for given vallue of pole positions poles
    Args:
        poles           :   pole positions
        GM_matrix       :   Exact Matsubara Green's function data in the shape
                            (num_imag, num_orb, num_orb)
        zM              :   Matsubara frequencies
    Returns:
        grad            :   Gradient of E with respect to poles
    """

    #
    # Initialization
    #

    # number of imag frequency points
    num_imag = len(zM)
    # Get the number of poles.
    Num_poles = len(poles)

    # Get the norm of obj.
    err_value, X_vec, G_approx = cvx_optimize(poles, GM_matrix, zM)
    Gdiff = GM_matrix - G_approx

    # compute gradient
    grad = np.zeros((Num_poles))

    # Loop over all Matsubara frequencies
    for iw in range(num_imag):
        # Initialize the variable Ghere.
        dG_iw = Gdiff[iw]

        # Loop over all poles.
        for k in range(Num_poles):
            # Add the contribution from the pole k to grad.
            gradhere = X_vec[k, :, :] / (
                (1j * zM[iw] - poles[k])**2
            )
            grad[k] += np.real(np.einsum('ij, ij ->', dG_iw, gradhere))

    # Scale the gradient by -1/y.
    grad = -grad / err_value

    return grad


def run_es(
    iw_vals, G_iw, re_w_vals, diag=True, eta=0.01, eps_pol=1,
    ofile='Xw_real.txt', solver='SCS', **kwargs
):
    """Pole estimation and semi-definite relaation algorithm based on
    Huang, Gull and Lin, 10.1103/PhysRevB.107.075151
    Input args
        iw_vals     :   imaginary frequency values (value only, without iota)
        G_iw        :   matsubara quantity to be analytically continued
        re_w_vals   :   real frequency grid to perform continuation on
        diag        :   True / False decides how to encode the convex problem
        eta         :   broadening
        eps_pol     :   max imag part of the poles to be considered
        ofile       :   Name of output file for storing the continued data
        solver      :   CVXPy solver to use
                        (default is SCS; other options: MOSEK, CLARABEL)
        **kwargs    :   other options specific to the CVXPy solver
    """

    # step 1: AAA algorithm for pole estimation
    if len(G_iw.shape) == 1:
        G_iw = G_iw.reshape(G_iw.shape + (1, ))
        g_trace = G_iw * 1.0
    elif len(G_iw.shape) == 2:
        g_trace = np.einsum('wa -> w', G_iw)
    elif len(G_iw.shape) == 3:
        g_trace = np.einsum('wpp -> w', G_iw)
    else:
        raise ValueError(
            'expect X_iw shape to be (nw, nao) or (nw, nao, nao), '
            + 'but received {}'.format(G_iw.shape)
        )
    res = baryrat.aaa(1j * iw_vals, g_trace)
    poles, _ = res.polres()
    poles = poles[np.abs(np.imag(poles)) < eps_pol]
    poles = np.asarray(np.real(poles))
    poles = np.unique(np.sort(poles))

    # step 2: semi-definite relaxation
    # optimize the poles
    if diag:
        res = minimize(
            lambda x: cvx_optimize_spectral(
                x, G_iw, iw_vals, solver, **kwargs
            )[0], poles, tol=1e-6, options={
                "maxiter": 30,
                "eps": 1e-7,
            }
        )
        poles_opt = res.x
        opt_error, X_vec, _ = cvx_optimize_spectral(
            poles_opt, G_iw, iw_vals, solver, **kwargs
        )
    else:
        res = minimize(
            lambda x: cvx_optimize(x, G_iw, iw_vals, solver, **kwargs)[0],
            poles, jac=lambda x: cvx_gradient(x, G_iw, iw_vals),
            tol=1e-6, options={
                "maxiter": 30,
                "eps": 1e-7,
            }
        )
        poles_opt = res.x
        opt_error, X_vec, _ = cvx_optimize(
            poles_opt, G_iw, iw_vals, solver, **kwargs
        )

    # Calculate the spectrum
    Greens_calc = np.zeros(
        (len(re_w_vals), ) + G_iw.shape[1:], dtype=complex
    )
    for i_re, re_w in enumerate(re_w_vals):
        Greenhere = np.zeros(G_iw.shape[1:], dtype=complex)
        for i_pol in range(len(poles_opt)):
            Greenhere += X_vec[i_pol] / (
                re_w + eta * 1j - poles_opt[i_pol]
            )
        Greens_calc[i_re] = Greenhere

    np.savetxt(ofile, Greens_calc.reshape(len(re_w_vals), -1))
    np.savetxt(ofile.replace('.txt', '_error.txt'), np.asarray([opt_error]))

    return





class Namespace:
    '''
    Generic namespace class that can be set by a dictionary object
    and return None for missing arguments
    '''
    def __init__(self, params):
        for name in params:
            setattr(self, name, params[name])

    def __eq__(self, other):
        if not isinstance(other, Namespace):
            return NotImplemented
        return vars(self) == vars(other)

    def __contains__(self, key):
        return key in self.__dict__

    def __getattr__(self, name):
        return None



import os

import h5py
import logging
import numpy as np
from pyscf.df import addons
from pyscf.pbc import tools, gto

from . import gdf_s_metric as gdf_S
from . import common_utils as comm
from . import integral_utils as int_utils



class pyscf_init:
    '''
    Attributes
    ----------
    args : map
        argument's map
    cell : pyscf.pbc.cell
        unit cell object
    kmesh : numpy.ndarray
        Monkhorst-Pack reciprocal space grid
    '''

    def __init__(self, args):
        '''
        Initialize PySCF interoperation class

        Parameters
        ----------
        args: map
            argument's map
        '''
        self.args = args
        if self.args.Nk is None:
            self.args.Nk = 0
        if self.args.spin is None:
            self.args.spin = 0
        if self.args.damping is None:
            self.args.damping = 0
        if self.args.max_iter is None:
            self.args.max_iter = 100
        self.cell = self.cell_object()
        
    
    def compute_df_int(self, nao, X_k):
        raise NotImplementedError("Please Implement this method")
    def mf_object(self, mydf=None):
        raise NotImplementedError("Please Implement this method")
    def df_object(self, mydf=None):
        raise NotImplementedError("Please Implement this method")
    def cell_object(self):
        raise NotImplementedError("Please Implement this method")
    def mean_field_input(self, mydf=None):
        raise NotImplementedError("Please Implement this method")



class pyscf_pbc_init (pyscf_init):
    '''
    '''
    def __init__(self, args=None):
        super().__init__(comm.init_pbc_params() if args is None else args)
        self.kmesh, self.k_ibz, self.ir_list, self.conj_list, self.weight, self.ind, self.num_ik = comm.init_k_mesh(self.args, self.cell)

    def mean_field_input(self, mydf=None):
        '''
        Solve a give mean-field problem and store the solution in the Green/WeakCoupling format
        
        Parameters
        ----------
        mydf : pyscf.pbc.df
            pyscf density-fitting object, will be generated if None
        '''
        if mydf is None:
            mydf = self.df_object()

        if os.path.exists("cderi.h5"):
            mydf._cderi = "cderi.h5"
        else:
            mydf._cderi_to_save = "cderi.h5"
            mydf.build()
        # number of k-points in each direction for Coulomb integrals
        nk       = self.args.nk ** 3
        # number of k-points in each direction to evaluate Coulomb kernel
        Nk       = self.args.Nk

        # number of orbitals per cell
        nao = self.cell.nao_nr()
        Zs = np.asarray(self.cell.atom_charges())
        logging.info(f"Number of atoms: {Zs.shape[0]}")
        logging.info("Effective nuclear charge of each atom: {Zs")
        atoms_info = np.asarray(self.cell.aoslice_by_atom())
        last_ao = atoms_info[:,3]
        logging.info(f"aoslice_by_atom = {atoms_info}")
        logging.info(f"Last AO index for each atom = {last_ao}")

        if self.args.grid_only:
            comm.store_k_grid(self.args, self.cell, self.kmesh, self.k_ibz, self.ir_list, self.conj_list, self.weight, self.ind, self.num_ik)
            return

        '''
        Generate integrals for mean-field calculations
        '''
        auxcell = addons.make_auxmol(self.cell, mydf.auxbasis)
        NQ = auxcell.nao_nr()
    
        mf = self.mf_object(mydf)
    
        # Get Overlap and Fock matrices
        hf_dm = mf.make_rdm1().astype(dtype=np.complex128)
        S     = mf.get_ovlp().astype(dtype=np.complex128)
        T     = mf.get_hcore().astype(dtype=np.complex128)
        if self.args.xc is not None:
            vhf = mf.get_veff().astype(dtype=np.complex128)
        else:
            vhf = mf.get_veff(hf_dm).astype(dtype=np.complex128)
        F = mf.get_fock(T,S,vhf,hf_dm).astype(dtype=np.complex128)
    
        if len(F.shape) == 3:
            F     = F.reshape((1,) + F.shape)
            hf_dm = hf_dm.reshape((1,) + hf_dm.shape)
        S = np.array((S, ) * self.args.ns)
        T = np.array((T, ) * self.args.ns)
    
        X_k = []
        X_inv_k = []

        # Orthogonalization matrix
        X_k, X_inv_k, S, F, T, hf_dm = comm.orthogonalize(mydf, self.args.orth, X_k, X_inv_k, F, T, hf_dm, S)
        comm.save_data(self.args, self.cell, mf, self.kmesh, self.ind, self.weight, self.num_ik, self.ir_list, self.conj_list, Nk, nk, NQ, F, S, T, hf_dm, Zs, last_ao)
        if bool(self.args.df_int) :
            self.compute_df_int(nao, X_k)

    def compute_df_int(self, nao, X_k):
        '''
        Generate density-fitting integrals for correlated methods
        '''
        mydf = comm.construct_gdf(self.args, self.cell, self.kmesh)
        int_utils.compute_integrals(self.args, self.cell, mydf, self.kmesh, nao, X_k, "df_hf_int", "cderi.h5", True, self.args.keep_cderi)
        mydf = None

        if self.args.finite_size_kind in ['gf2', 'gw', 'gw_s'] :
            self.compute_twobody_finitesize_correction()
            return

        mydf = comm.construct_gdf(self.args, self.cell, self.kmesh)
        # Use Ewald for divergence treatment
        mydf.exxdiv = 'ewald'
        import importlib.util as iu
        new_pyscf = iu.find_spec('pyscf.pbc.df.gdf_builder') is not None
        if new_pyscf :
            import pyscf.pbc.df.gdf_builder as gdf
            weighted_coulG_old = gdf._CCGDFBuilder.weighted_coulG
            gdf._CCGDFBuilder.weighted_coulG = int_utils.weighted_coulG_ewald
        else:
            from pyscf.pbc import df as gdf
            weighted_coulG_old = gdf.GDF.weighted_coulG
            gdf.GDF.weighted_coulG = int_utils.weighted_coulG_ewald
    
        #kij_conj, kij_trans, kpair_irre_list, kptij_idx, num_kpair_stored = 
        int_utils.compute_integrals(self.args, self.cell, mydf, self.kmesh, nao, X_k, "df_int", "cderi_ewald.h5", True, self.args.keep_cderi)
        if new_pyscf :
            gdf._CCGDFBuilder.weighted_coulG = weighted_coulG_old
        else:
            gdf.GDF.weighted_coulG = weighted_coulG_old

    def evaluate_high_symmetry_path(self):
        if self.args.print_high_symmetry_points:
            comm.print_high_symmetry_points(self.cell, self.args)
            return
        if self.args.high_symmetry_path is None:
            raise RuntimeError("Please specify high-symmetry path")
        if self.args.high_symmetry_path is not None:
            try:
                comm.check_high_symmetry_path(self.cell, self.args)
            except RuntimeError as e:
                logging.error("\n\n\n")
                logging.error("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
                logging.error("!!!!!!!!! Cannot compute high-symmetry path !!!!!!!!!")
                logging.error("!! Correct or Disable high-symmetry path evaluation !")
                logging.error("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
                logging.error(e)
                exit(-1)
        kmesh_hs, Hk_hs, Sk_hs = comm.high_symmetry_path(self.cell, self.args)
        inp_data = h5py.File(self.args.output_path, "a")
        logging.debug(kmesh_hs)
        logging.debug(self.cell.get_scaled_kpts(kmesh_hs))
        inp_data["high_symm_path/k_mesh"] = self.cell.get_scaled_kpts(kmesh_hs)
        inp_data["high_symm_path/r_mesh"] = comm.construct_rmesh(self.args.nk, self.args.nk, self.args.nk)
        inp_data["high_symm_path/Hk"] = Hk_hs
        inp_data["high_symm_path/Sk"] = Sk_hs

    def compute_twobody_finitesize_correction(self, mydf=None):
        if not os.path.exists(self.args.hf_int_path):
            os.mkdir(self.args.hf_int_path)
        if self.args.finite_size_kind == 'gf2' :
            comm.compute_ewald_correction(self.args, self.cell, self.kmesh, self.args.hf_int_path + "/df_ewald.h5")
        elif self.args.finite_size_kind == 'gw' :
            self.evaluate_gw_correction(mydf)
            
    
    def evaluate_gw_correction(self, mydf=None):
        if mydf is None:
            mydf = comm.construct_gdf(self.args, self.cell, self.kmesh)
        mydf.build()

        j3c, kptij_lst, j2c_sqrt, uniq_kpts = gdf_S.make_j3c(mydf, self.cell, j2c_sqrt=True, exx=False)
        
        ''' Transformation matrix from auxiliary basis to plane-wave '''
        AqQ, q_reduced, q_scaled_reduced = gdf_S.transformation_PW_to_auxbasis(mydf, self.cell, j2c_sqrt, uniq_kpts)
        
        q_abs = np.array([np.linalg.norm(qq) for qq in q_reduced])
        q_abs = np.array([round(qq, 8) for qq in q_abs])
        
        # Different prefactors for the GW finite-size correction for testing
        # In practice, the madelung constant is used, which decays as (1/nk).
        X = (6*np.pi**2)/(self.cell.vol*len(self.kmesh))
        X = (2.0/np.pi) * np.cbrt(X)
        
        X2 = 2.0 * np.cbrt(1.0/(self.cell.vol*len(self.kmesh)))
        
        f = h5py.File(self.args.hf_int_path + "/AqQ.h5", 'w')
        f["AqQ"] = AqQ
        f["qs"] = q_reduced
        f["qs_scaled"] = q_scaled_reduced
        f["q_abs"] = q_abs
        f["X"] = X
        f["X2"] = X2
        f["madelung"] = tools.pbc.madelung(self.cell, self.kmesh)
        f.close()

    def mf_object(self, mydf=None):
        return comm.solve_mean_field(self.args, mydf, self.cell)

    def df_object(self, mydf=None):
        return comm.construct_gdf(self.args, self.cell, self.kmesh)

    def cell_object(self):
        return comm.pbc_cell(self.args)

class pyscf_mol_init (pyscf_init):
    '''
    '''
    def __init__(self, args=None):
        super().__init__(comm.init_mol_params() if args is None else args)
        self.kmesh = np.array([[0.,0.,0.]])
        self.k_ibz = np.array([[0.,0.,0.],])
        self.ir_list = np.array([0])
        self.conj_list= np.array([0])
        self.weight= np.array([1.0])
        self.ind= np.array([0])
        self.num_ik = 1
        self.kcell = gto.Cell(verbose=0)
        self.kcell.a = [[1,0,0],[0,1,0],[0,0,1]]
        self.kcell.atom = self.cell.atom
        self.kcell.spin = self.cell.spin
        self.kcell.charge = self.cell.charge
        self.kcell.unit = 'A'
        self.kcell.basis = self.cell.basis
        self.kcell.kpts = self.kcell.make_kpts([1, 1, 1])
        self.kcell.ecp = self.cell.ecp
        self.kcell.build()

    def mean_field_input(self, mydf=None):
        '''
        Solve a give mean-field problem and store the solution in the Green/WeakCoupling format
        
        Parameters
        ----------
        mydf : pyscf.df
            pyscf density-fitting object, will be generated if None
        '''
        if mydf is None:
            mydf = self.df_object()
#comm.construct_gdf(self.args, self.cell, self.kmesh)

        # number of k-points in each direction for Coulomb integrals
        nk       = self.args.nk ** 3
        # number of k-points in each direction to evaluate Coulomb kernel
        Nk       = self.args.Nk

        # number of orbitals per cell
        nao = self.cell.nao_nr()
        Zs = np.asarray(self.cell.atom_charges())
        logging.info(f"Number of atoms: {Zs.shape[0]}")
        logging.info("Effective nuclear charge of each atom: {Zs")
        atoms_info = np.asarray(self.cell.aoslice_by_atom())
        last_ao = atoms_info[:,3]
        logging.info(f"aoslice_by_atom = {atoms_info}")
        logging.info(f"Last AO index for each atom = {last_ao}")

        '''
        Generate integrals for mean-field calculations
        '''
        auxcell = addons.make_auxmol(self.cell, mydf.auxbasis)
        NQ = auxcell.nao_nr()
    
        mf = self.mf_object(mydf)
    
        # Get Overlap and Fock matrices
        hf_dm = mf.make_rdm1().astype(dtype=np.complex128)
        S     = mf.get_ovlp().astype(dtype=np.complex128)
        T     = mf.get_hcore().astype(dtype=np.complex128)
        if self.args.xc is not None:
            vhf = mf.get_veff().astype(dtype=np.complex128)
        else:
            vhf = mf.get_veff(hf_dm).astype(dtype=np.complex128)
        F = mf.get_fock(T,S,vhf,hf_dm).astype(dtype=np.complex128)


        F = F.reshape((self.args.ns, self.args.nk, nao, nao))
        hf_dm = hf_dm.reshape((self.args.ns, self.args.nk, nao, nao))
        S = S.reshape((self.args.nk, nao, nao))
        T = T.reshape((self.args.nk, nao, nao))
    
        if len(F.shape) == 3:
            F     = F.reshape((1,) + F.shape)
            hf_dm = hf_dm.reshape((1,) + hf_dm.shape)
        S = np.array((S, ) * self.args.ns)
        T = np.array((T, ) * self.args.ns)

        X_k = []
        X_inv_k = []

        # Orthogonalization matrix
        X_k, X_inv_k, S, F, T, hf_dm = comm.orthogonalize(mydf, self.args.orth, X_k, X_inv_k, F, T, hf_dm, S)
        comm.save_data(self.args, self.kcell, mf, self.kmesh, self.ind, self.weight, self.num_ik, self.ir_list, self.conj_list, Nk, nk, NQ, F, S, T, hf_dm, Zs, last_ao)
        if bool(self.args.df_int):
            self.compute_df_int(nao, X_k)

    def compute_df_int(self, nao, X_k):
        '''
        Generate density-fitting integrals for correlated methods
        '''
        h_in = h5py.File("cderi_mol.h5", 'a')
        h_out = h5py.File("cderi.h5", 'w')

        j3c_obj = h_in["/j3c"]
        if not isinstance(j3c_obj, h5py.Dataset):  # not a dataset
            if isinstance(j3c_obj, h5py.Group):  # pyscf >= 2.1
                h_in.copy(h_in["/j3c"], h_out, "j3c/0")
            else:
                raise ValueError("Unknown structure of cderi_mol.h5. Perhaps, PySCF upgrade went badly...")
        else:  # pyscf < 2.1
            h_in.copy(h_in["/j3c"], h_out, "j3c/0/0")

        kptij = np.zeros((1, 2, 3))
        h_out["j3c-kptij"] = kptij

        h_in.close()
        h_out.close()
        mydf = comm.construct_gdf(self.args, self.kcell, self.kmesh)
        int_utils.compute_integrals(self.args, self.kcell, mydf, self.kmesh, nao, X_k, "df_hf_int", "cderi.h5", True, self.args.keep_cderi)
        mydf = None

    def df_object(self, mydf=None):
        return comm.construct_mol_gdf(self.args, self.kcell)

    def mf_object(self, mydf=None):
        return comm.solve_mol_mean_field(self.args, mydf, self.cell)

    def cell_object(self):
        return comm.mol_cell(self.args)



import numpy as np
from . import analyt_cont as ac

"""
Perform quasiparticle approximation
"""

"""
Construct quasiparticle renormalization factors in a local Hamiltonian
F = (ns, nao, nao)
Sigma_iw = (nw, ns, nao, nao)
"""


def Z_factor(
    F, Sigma_iw, iwsample, outdir='sigma_nevan',
    ifile='Sigma_iw.txt', ofile='Sigma_w.txt', coefile='coeff',
    n_real=10001, w_min=-10, w_max=10, eta=0.01
):
    ns, nao = F.shape[:2]
    # Nevanlinna continuation of self-energy.
    Sigma_iw_diag = np.einsum('wsii->wsi', Sigma_iw)
    print("Sigma_iw_diag shape = {}".format(Sigma_iw_diag.shape))
    nw = iwsample.shape[0]
    iwsample_pos = iwsample[nw//2:]
    Sigma_iw_diag = Sigma_iw_diag[nw//2:]
    freqs, Sigma_w = ac.nevan_run(
        Sigma_iw_diag, iwsample_pos, outdir=outdir, ifile=ifile, ofile=ofile,
        coeff_file=coefile, n_real=n_real, w_min=w_min, w_max=w_max, eta=eta,
        green=False
    )

    # F_diag = (ns, nno)
    F_diag = np.array([np.diag(f) for f in F])
    dw = freqs[1] - freqs[0]
    Zs = np.zeros((ns, nao))
    for s in range(ns):
        for i in range(nao):
            e_approx, idx = find_nearest(freqs, F_diag[s, i])
            # Z = [1 - d(Sigma)/d(w)|w=e]^{-1}
            # Z^{-1} >= 1
            Z = 1 - np.gradient(Sigma_w[:, s, i].real, dw)
            Z = 1 / Z
            Zs[s, i] = Z[idx]

    return Zs


def find_nearest(array, value):
    import math
    idx = np.searchsorted(array, value, side="left")
    cond1 = idx > 0
    cond2 = idx == len(array) or \
        math.fabs(value - array[idx - 1]) < math.fabs(value - array[idx])

    if cond1 and cond2:
        return array[idx-1], idx-1
    else:
        return array[idx], idx



import os
import subprocess
import shutil
import numpy as np
import h5py
from multiprocessing import cpu_count, Process
from .pes_utils import run_es, cvx_matrix_projection, cvx_diag_projection
from .ac_utils import dump_input_caratheodory_data, load_caratheodory_data
import green_mbtools.pesto.nevanlinna as nevan_exe
import green_mbtools.pesto.caratheodory as carath_exe


# Find the number of cpus to use for parallelization of analtyic continuation
slurm_env_var = 'SLURM_JOB_CPUS_PER_NODE'
phys_ncpu = cpu_count()

if slurm_env_var in os.environ:
    _ncpu = int(
        os.environ['SLURM_JOB_CPUS_PER_NODE']
    )
else:
    print(
        "Variable SLURM_JOB_CPUS_PER_NODE not found. "
        + "Using half of the physical number of threads available."
    )
    _ncpu = int(phys_ncpu // 2)


def maxent_run(
    gtau, tau_mesh, error=5e-3, params="green.param", exe_path='maxent',
    outdir="Maxent"
):
    """Maxent analytic continuation for G(iw) or diagonal of self-energy

    Input parameters
    ----------------
        gtau        :   contains matrix valued data on imaginary time axis
                        shape of gtau should be:  (nt, :)
        tau_mesh    :   tau grid points
        error       :   error threshold in optimization
        params      :   parameter file name to pass into maxent continuation
        exe_path    :   Path to maxent program provided by CQMP:
                        https://github.com/CQMP/Maxent
        outdir      :   output directory in which data will be stored

    Returns
    ----------------
        The maxent spectral function is stored in '<outdir>/DOS.h5'
    """

    wkdir = os.path.abspath(os.getcwd())
    # outdir = os.path.abspath(wkdir+'/'+outdir)
    print("Maxent output:", os.path.abspath(wkdir+'/'+outdir))

    beta = tau_mesh[-1]
    nts = tau_mesh.shape[0]
    g_shape = gtau.shape
    assert nts == gtau.shape[0], "Number of imaginary time points mismatches."
    gtau = gtau.reshape(nts, -1)
    dim1 = gtau.shape[1]

    if not os.path.exists(outdir):
        os.mkdir(outdir)
    os.chdir(outdir)

    # Save dim1 for better understanding of output
    np.savetxt("dimensions.txt", np.asarray(g_shape[1:], dtype=int))
    # FIXME Can we have multiple layers of folders so that one can separate
    # the whole job into chunks?
    for d1 in range(dim1):
        if not os.path.exists(str(d1)):
            os.mkdir(str(d1))
        np.savetxt(
            "{}/G_tau.txt".format(d1),
            np.column_stack(
                (tau_mesh, gtau[:, d1].real, np.array([error] * nts))
            )
        )

    # Start analytical continuation
    processes = []
    pp = 0
    for d1 in range(dim1):
        os.chdir(str(d1))
        try:
            shutil.copy(str(params), "./green.param")
        except Exception:
            shutil.copy(os.path.abspath(wkdir+'/'+params), "./green.param")

        with open("log.txt", "w") as log:
            p = subprocess.Popen(
                [
                    exe_path, "./green.param", "--DATA=G_tau.txt",
                    "--BETA=" + str(beta), "--NDAT=" + str(nts)
                ],
                stdout=log, stderr=log
            )
            processes.append(p)
        pp += 1
        if pp % _ncpu == 0:
            for proc in processes:
                proc.communicate()
            processes = []
        os.chdir("..")

    for proc in processes:
        proc.communicate()

    # Combine output
    dump_A = False
    for d1 in range(dim1):
        try:
            freqs = np.loadtxt("{}/green.out.maxspec.dat".format(d1))[:, 0]
        except IOError:
            pass
        else:
            dump_A = True
            break
    if dump_A:
        Aw = np.zeros((freqs.shape[0], dim1), dtype=float)
        for d1 in range(dim1):
            try:
                Aw[:, d1] = np.loadtxt(
                    "{}/green.out.maxspec.dat".format(d1)
                )[:, 1]
            except IOError:
                print(
                    "green.out.maxspec.dat is missing in {} folder. Possibly \
                    analytical continuation fails at that point.".format(d1)
                )
        Aw = Aw.reshape((freqs.shape[0],) + g_shape[1:])
        gtau = gtau.reshape(g_shape)
        f = h5py.File("DOS.h5", 'w')
        f["freqs"] = freqs
        f["DOS"] = Aw
        f["taumesh"] = tau_mesh
        f["Gtau"] = gtau
        f.close()
    else:
        print("All AC fails. Will not dump to DOS.h5")
        gtau = gtau.reshape(g_shape)

    os.chdir("..")


def nevan_run(
    X_iw, wsample, outdir='Nevanlinna', n_real=10000, w_min=-10, w_max=10,
    eta=0.01, spectral=True, prec=128
):
    """Nevanlinna analytic continuation for G(iw) or diagonal of self-energy
    Input parameters
    ----------------
        X_iw        :   contains matrix valued data on imaginary freq. points
                        shape of X_iw should be:  (nw, ns, nk, nao, nao)
        wsample     :   value of imaginary frequencies
        outdir      :   output directory in which data will be stored
        n_real      :   number of real frequency points
                        (used if custom_freqs = None)
        w_min       :   minimum value of real frequency range
                        (used if custom_freqs = None)
        w_max       :   maximum value of real frequency range
                        (used if custom_freqs = None)
        eta         :   broadening parameter
        spectral    :   returns real valued spectral function if True,
                        full complex valued function otherwise
        prec        :   Precision to use in Nevanlinna

    Returns:
    ----------------
        freqs       :   Real frequency grid on which AC data is obtained
        X_w         :   Real valued spectral function (if green=False)
                        or complex valued AC quantity (if green=True)
    """
    # print acknowledgments
    print("----------------------------------------------")
    print("Performing Nevanlinna analytic continuation.")
    print("Reference:")
    print("Fei et al, Phys. Rev. Lett. 126, 056402 (2021)")
    print("----------------------------------------------")

    # define defaults
    ifile = 'X_iw.txt'
    ofile = 'X_w.txt'
    coeff_file = 'coeff.txt'
    wkdir = os.path.abspath(os.getcwd())
    print("Nevanlinna output:", os.path.abspath(wkdir + '/' + outdir))
    print("Will dump input to {} and output to {}".format(ifile, ofile))

    nw = wsample.shape[0]
    assert nw == X_iw.shape[0], "Number of imaginary frequency points \
        mismatches between \"input_parser\" and Gw."
    X_iw_shape = X_iw.shape
    X_iw = X_iw.reshape(nw, -1)
    dim1 = X_iw.shape[1]

    if not os.path.exists(outdir):
        os.mkdir(outdir)
    os.chdir(outdir)

    # Save dim1 for better understanding of output
    np.savetxt("dimensions.txt", np.asarray(X_iw_shape[1:], dtype=int))

    for d1 in range(dim1):
        if not os.path.exists(str(d1)):
            os.mkdir(str(d1))
        np.savetxt(
            "{}/{}".format(d1, ifile),
            np.column_stack(
                (wsample, X_iw[:, d1].real, X_iw[:, d1].imag)
            )
        )

    # Start analytical continuation
    processes = []
    pp = 0
    for d1 in range(dim1):
        os.chdir(str(d1))
        p = Process(
            target=nevan_exe.nevanlinna,
            args=(
                ifile, nw, ofile, coeff_file, prec, spectral,
                n_real, w_min, w_max, eta
            )
        )
        p.start()
        processes.append(p)
        pp += 1
        if pp % _ncpu == 0:
            for proc in processes:
                proc.join()
            processes = []
        os.chdir("..")

    for proc in processes:
        proc.join()

    # Combine output
    dump_A = False
    try:
        X_w = np.loadtxt("0/{}".format(ofile))
        dtype = complex if X_w.shape[1] == 3 else float
        freqs = X_w[:, 0]
        X_wk = np.loadtxt("0/{}".format(ofile))
        dtype = complex if X_wk.shape[1] == 3 else float
        freqs = X_wk[:, 0]
    except IOError:
        pass
    else:
        dump_A = True
    if dump_A:
        X_w = np.zeros((freqs.shape[0], dim1), dtype=dtype)
        for d1 in range(dim1):
            # Read X_w data
            try:
                X_wk = np.loadtxt("{}/{}".format(d1, ofile))
                if dtype == complex:
                    X_w[:, d1].real = X_wk[:, 1]
                    X_w[:, d1].imag = X_wk[:, 2]
                else:
                    X_w[:, d1] = X_wk[:, 1]
            except IOError:
                print(
                    "{} is missing in {} folder. Possibly analytical \
                    continuation fails at that point.".format(ofile, d1)
                )
        X_w = X_w.reshape((freqs.shape[0],) + X_iw_shape[1:])
        X_iw = X_iw.reshape(X_iw_shape)
    else:
        print("All AC fails. Will not dump to DOS.h5")
        X_iw = X_iw.reshape(X_iw_shape)

    os.chdir(wkdir)

    return freqs, X_w


def caratheodory_run(
    X_iw, wsample, outdir='Caratheodory', custom_freqs=None,
    n_real=2001, w_min=-10, w_max=10, eta=0.01,
):
    """Caratheodory analytic continuation for G(iw) or self-energy

    Input parameters
    ----------------
        X_iw            : contains matrix valued data on imaginary freq. points
                        shape of X_iw should be: (nw, ns, nk, nao, nao)
        wsample         : value of imaginary frequencies
        outdir          : output directory in which data will be stored
        custom_freqs    : custom freq points on which to perform carath. AC
        n_real          : number of real frequency points
                          (used if custom_freqs = None)
        w_min           : minimum value of real frequency range
                          (used if custom_freqs = None)
        w_max           : maximum value of real frequency range
                          (used if custom_freqs = None)
        eta             : broadening parameter

    Returns:
    ----------------
        freqs       :   Real frequency grid on which AC data is obtained
        Xc_w        :   analytically continued matrix data
        XA_w        :   spectral function for AC data
    """

    # print acknowledgments
    print("----------------------------------------------")
    print("Performing Caratheodory analytic continuation.")
    print("Reference:")
    print("Fei et al, Phys. Rev. B 104, 165111 (2021)")
    print("----------------------------------------------")

    # set defaults
    ifile = 'X_iw.txt'
    matrix_ofile = 'X_c.txt'
    spectral_ofile = 'X_A.txt'

    # create input data
    wkdir = os.path.abspath(os.getcwd())
    print("Caratheodory output:", os.path.abspath(wkdir + '/' + outdir))
    print("Dumping input to: ", ifile)
    print("Dumping Caratheodory matrix output to: ", matrix_ofile)
    print("Dumping Caratheodory spectral output to: ", spectral_ofile)

    nw = wsample.shape[0]
    assert nw == X_iw.shape[0], "Number of imaginary frequency points \
        mismatches between \"input_parser\" and Gw."

    # whether or not to use custom grid for continuation
    use_custom_real_grid = 0  # False
    grid_file = 'grid_file.txt'
    if custom_freqs is not None:
        use_custom_real_grid = 1  # True
        n_real = custom_freqs.shape[0]

    # assuming that the input
    X_iw_shape = X_iw.shape
    _, ns, nk, nao = X_iw_shape[:4]

    # Build and move to working directory
    if not os.path.exists(outdir):
        os.mkdir(outdir)
    os.chdir(outdir)

    # Dump input data to files for caratheodory
    dump_input_caratheodory_data(wsample, X_iw, ifile)

    # Start analytical continuation
    processes = []
    pp = 0
    dim1 = ns * nk
    for d1 in range(dim1):
        os.chdir(str(d1))
        if custom_freqs is not None:
            np.savetxt(grid_file, custom_freqs)
        p = Process(
            target=carath_exe.caratheodory,
            args=(
                ifile, nw, nao, matrix_ofile, spectral_ofile,
                use_custom_real_grid, grid_file, n_real, w_min, w_max, eta
            )
        )
        p.start()
        processes.append(p)
        pp += 1
        if pp % _ncpu == 0:
            for proc in processes:
                proc.join()
            processes = []
        os.chdir("..")

    for proc in processes:
        proc.join()

    # Combine output
    freqs, Xc_w, XA_w = load_caratheodory_data(
        matrix_ofile, spectral_ofile, X_iw_shape
    )

    # Get out of working directory
    os.chdir(wkdir)

    return freqs, Xc_w, XA_w


def es_nevan_run(
    G_iw, wsample, n_real=10000, w_min=-10, w_max=10, eta=0.01, diag=True,
    eps_pol=1.0, parallel='sk', outdir='PESNevan', solver='SCS', **solver_opts
):
    """ES analytic continuation for G(iw) or diagonal of self-energy.

    Input parameters
    ----------------
        G_iw        :   imaginary frequency green's function
        wsample     :   (real part) of the imaginary frequencies
        n_real      :   number of grid-points to use on the real freuency axis
        w_min       :   lowest real frequency to get results for
        w_max       :   highest real frequency to get results for
        eta         :   broadening for real-axis.
        diag        :   value of True will perform analytic continuation only
                        for the diagonal values of Green's function
                        (precisely what we need for spectral function)
                        (NOTE: we can diagonalize the G_iw first then input
                        with diag=False as well)
        eps_pol     :   Threshold on the imaginary part of poles
        parallel    :   How to parallelize (over openMP) the analytic cont.
                        'sk' meanse we parallelize over spin and k-ponts.
                        'ska' would parallelize over spin, k and orbitals
                        Typically, 'sk' is most optimal, because the
                        analyic continuation is vectorized over orbtal indices
        solver      :   ES uses semi-definite programming to perform
                        analytical continuation. Different solvers can be
                        employed, e.g., 'SCS', 'MOSEK', 'CLARABEL'.
        solver_opts :   A dictionary of options, e.g., tolerance, can be
                        passed to this function as well

    Returns
    ----------------
        w_vals      :   real frequency grid
        G_w         :   analytically continued Green's function on real grid
    """

    # print acknowledgments
    print("----------------------------------------------")
    print("Performing ES analytic continuation.")
    print("Reference:")
    print("Huang et al, Phys. Rev. B 107, 075151 (2023)")
    print("----------------------------------------------")

    # set defaults
    ofile = 'Aw.txt'

    # Handle directories for saving output files
    wkdir = os.path.abspath(os.getcwd())
    print("PES Nevanlinna output:", os.path.abspath(wkdir + '/' + outdir))
    if not os.path.exists(outdir):
        os.mkdir(outdir)
    base_dir = wkdir + '/' + outdir

    # Bring G_iw in the shape: (iw, s, k, a, b)
    orig_shape = G_iw.shape
    if len(orig_shape) == 1:
        # this is simply G(iw)
        G_iw = G_iw.reshape((orig_shape[0], 1, 1, 1))
        print("Expecting dimensions to represent (w)")
    elif len(orig_shape) == 2:
        # assuming this is G(iw, a)
        nw, nao = orig_shape
        G_iw = G_iw.reshape((nw, 1, 1, nao))
        print("Expecting dimensions to represent (w, a)")
    elif len(orig_shape) == 4:
        # for diagonal approx only
        print("Expecting dimensions to represent (w, s, k, a)")
    elif len(orig_shape) == 5:
        # this is already the form we want
        print("Expecting dimensions to represent (w, s, k, a, b)")
    else:
        raise ValueError(
            'Incorrect shape G_iw. Expecting either 1-, 2-, 4- or 5-d array'
        )

    # Diagonal elements or not
    if diag and len(G_iw.shape) == 5:
        G_iw = np.einsum('wskaa -> wska', G_iw, optimize=True)
        # update the shape
        orig_shape = G_iw.shape

    # Check consistency in number of frequencies and input data
    nw, ns, nk = G_iw.shape[0:3]
    assert nw == wsample.shape[0], "Number of imaginary frequency points \
        mismatches between \"input_parser\" and Gw."

    # Pre-processing - divide the data based on k-, spin-, and AO-indices
    # for parallelized analytic continuation
    if parallel == 'ska':
        if not diag:
            raise ValueError(
                "Cannot use the 'ska' parallel scheme for full matrix"
            )
        dim1 = np.prod(G_iw.shape[1:])
        G_iw = G_iw.reshape(nw, dim1, 1)
    elif parallel == 'sk':
        dim1 = ns * nk
        G_iw = G_iw.reshape((nw, ns * nk) + G_iw.shape[3:])
    else:
        raise ValueError(
            "Invalid value for the argument 'parralel': {}.".format(parallel)
            + "Expecting 'ska' or 'sk'"
        )

    # Save dimensions for better understanding of output
    np.savetxt("dimensions.txt", np.asarray(G_iw.shape[1:], dtype=int))

    # define the frequency mesh
    w_vals = np.linspace(w_min, w_max, n_real)

    # perform analytical continuation
    processes = []
    pp = 0
    for d1 in range(dim1):
        if not os.path.exists(base_dir + '/' + str(d1)):
            os.mkdir(base_dir + '/' + str(d1))
        out_file = base_dir + '/{}/{}'.format(str(d1), ofile)
        p = Process(
            target=run_es,
            args=(
                wsample, G_iw[:, d1, :], w_vals
            ),
            kwargs={
                'diag': diag,
                'eta': eta,
                'eps_pol': eps_pol,
                'ofile': out_file,
                'solver': solver,
                **solver_opts
            }
        )
        p.start()
        processes.append(p)
        pp += 1
        if pp % _ncpu == 0:
            for proc in processes:
                proc.join()
            processes = []

    for proc in processes:
        proc.join()

    # read the computed spectrum and quantitie
    G_w = np.zeros((n_real, ) + G_iw.shape[1:], dtype=complex)
    error_es = np.zeros(dim1)
    pp = 0
    for d1 in range(dim1):
        # G_w output
        out_file = base_dir + '/{}/{}'.format(str(d1), ofile)
        Gw_here = np.loadtxt(out_file, dtype=complex)
        G_w[:, d1, :] = Gw_here.reshape((n_real, ) + G_iw.shape[2:])
        # G_w optimization error
        err_file = out_file.replace('.txt', '_error.txt')
        err_here = np.loadtxt(err_file)
        error_es[d1] = err_here

    # reshape and return
    G_w = G_w.reshape((n_real, ) + orig_shape[1:])
    error_es = error_es.reshape(orig_shape[1:])

    return w_vals, G_w, error_es


def g_iw_projection(
    G_iw, wsample, diag=True, solver='SCS', wcut=5, n_real=101, **solver_opts
):
    """Projection of G(iw) to a simple pole structure.

    Input parameters
    ----------------
        G_iw        :   imaginary frequency green's function
        wsample     :   (real part) of the imaginary frequencies
        diag        :   value of True will perform analytic continuation only
                        for the diagonal values of Green's function
                        (precisely what we need for spectral function)
                        (NOTE: we can diagonalize the G_iw first then input
                        with diag=False as well)
        solver      :   Projection uses semi-definite programming to perform
                        analytical continuation. Different solvers can be
                        employed, e.g., 'SCS', 'MOSEK', 'CLARABEL'.
        solver_opts :   A dictionary of options, e.g., tolerance, can be
                        passed to this function as well
        wcut        :   real frequency cutoff to employ in projection
        n_real      :   number of points to use in the range (-wcut, wcut)

    Returns
    ----------------
        G_iw        :   projected Matsubara Green's function

    Practical info
    ----------------
    For different practical applications, different CVXPy solvers may be
    required to achieve optimal results. Some options to consider:
    *   SCS (default)
    *   MOSEK
    *   CLARABEL
    For individual use, all three options are available free of use.
    """

    # print acknowledgments
    print("----------------------------------------------")
    print("Performing Projection of Matsubara Green's function.")
    print("Reference:")
    print("Huang et al, Phys. Rev. B 107, 075151 (2023)")
    print("----------------------------------------------")

    # Handle directories for saving output files
    outdir = 'Projection'
    ofile = 'Proj_Giw.txt'
    wkdir = os.path.abspath(os.getcwd())
    print("PES Nevanlinna output:", os.path.abspath(wkdir + '/' + outdir))
    if not os.path.exists(outdir):
        os.mkdir(outdir)
    base_dir = wkdir + '/' + outdir

    # Bring G_iw in the shape: (iw, s, k, a, b)
    orig_shape = G_iw.shape
    if len(orig_shape) == 1:
        # this is simply G(iw)
        G_iw = G_iw.reshape((orig_shape[0], 1, 1, 1))
        print("Expecting dimensions to represent (w)")
    elif len(orig_shape) == 2:
        # assuming this is G(iw, a)
        nw, nao = orig_shape
        G_iw = G_iw.reshape((nw, 1, 1, nao))
        print("Expecting dimensions to represent (w, a)")
    elif len(orig_shape) == 4:
        # for diagonal approx only
        print("Expecting dimensions to represent (w, s, k, a)")
    elif len(orig_shape) == 5:
        # this is already the form we want
        print("Expecting dimensions to represent (w, s, k, a, b)")
    else:
        raise ValueError(
            'Incorrect shape G_iw. Expecting either 1-, 2-, 4- or 5-d array'
        )

    # Diagonal elements or not
    projection_function = cvx_matrix_projection
    if diag:
        projection_function = cvx_diag_projection
    if diag and len(G_iw.shape) == 5:
        G_iw = np.einsum('wskaa -> wska', G_iw, optimize=True)
        # update the shape
        orig_shape = G_iw.shape

    # Check consistency in number of frequencies and input data
    nw, ns, nk, nao = G_iw.shape[0:4]
    assert nw == wsample.shape[0], "Number of imaginary frequency points \
        mismatches between \"input_parser\" and Gw."

    # Pre-processing - divide the data based on k-, spin--indices
    # for parallelized projection of data
    dim1 = ns * nk
    G_iw = G_iw.reshape((nw, ns * nk) + G_iw.shape[3:])

    # Save dimensions for better understanding of output
    np.savetxt("dimensions.txt", np.asarray(G_iw.shape[1:], dtype=int))

    # perform analytical continuation
    processes = []
    pp = 0
    for d1 in range(dim1):
        if not os.path.exists(base_dir + '/' + str(d1)):
            os.mkdir(base_dir + '/' + str(d1))
        out_file = base_dir + '/{}/{}'.format(str(d1), ofile)
        p = Process(
            target=projection_function,
            args=(
                1j * wsample, G_iw[:, d1, :]
            ),
            kwargs={
                'w_cut': wcut,
                'n_real': n_real,
                'ofile': out_file,
                'solver': solver,
                **solver_opts
            }
        )
        p.start()
        processes.append(p)
        pp += 1
        if pp % _ncpu == 0:
            for proc in processes:
                proc.join()
            processes = []

    for proc in processes:
        proc.join()

    # read the computed spectrum and quantitie
    if diag:
        G_iw_proj = np.zeros((nw, dim1, nao), dtype=complex)
    else:
        G_iw_proj = np.zeros((nw, dim1, nao * nao), dtype=complex)
    pp = 0
    for d1 in range(dim1):
        out_file = base_dir + '/{}/{}'.format(str(d1), ofile)
        Gw_here = np.loadtxt(out_file, dtype=complex)
        G_iw_proj[:, d1, :] = Gw_here * 1.0

    # reshape and return
    G_iw_proj = G_iw_proj.reshape((nw, ) + orig_shape[1:])

    return G_iw_proj



import logging
import os
import random
import string
import importlib.metadata as imd

import h5py
import numpy as np
from numba import jit
from pyscf.df import addons
from pyscf.pbc import df, tools


def compute_kG(k, Gv, wrap_around, mesh, cell):
    if abs(k).sum() > 1e-9:
        kG = k + Gv
    else:
        kG = Gv

    equal2boundary = np.zeros(Gv.shape[0], dtype=bool)
    if wrap_around and abs(k).sum() > 1e-9:
        # Here we 'wrap around' the high frequency k+G vectors into their lower
        # frequency counterparts.  Important if you want the gamma point and k-point
        # answers to agree
        b = cell.reciprocal_vectors()
        box_edge = np.einsum('i,ij->ij', np.asarray(mesh)//2+0.5, b)
        assert (all(np.linalg.solve(box_edge.T, k).round(9).astype(int)==0))
        reduced_coords = np.linalg.solve(box_edge.T, kG.T).T.round(9)
        on_edge = reduced_coords.astype(int)
        if cell.dimension >= 1:
            equal2boundary |= reduced_coords[:,0] == 1
            equal2boundary |= reduced_coords[:,0] ==-1
            kG[on_edge[:,0]== 1] -= 2 * box_edge[0]
            kG[on_edge[:,0]==-1] += 2 * box_edge[0]
        if cell.dimension >= 2:
            equal2boundary |= reduced_coords[:,1] == 1
            equal2boundary |= reduced_coords[:,1] ==-1
            kG[on_edge[:,1]== 1] -= 2 * box_edge[1]
            kG[on_edge[:,1]==-1] += 2 * box_edge[1]
        if cell.dimension == 3:
            equal2boundary |= reduced_coords[:,2] == 1
            equal2boundary |= reduced_coords[:,2] ==-1
            kG[on_edge[:,2]== 1] -= 2 * box_edge[2]
            kG[on_edge[:,2]==-1] += 2 * box_edge[2]
    return kG, equal2boundary



def get_coarsegrained_coulG(lattice_kmesh, cell, k=np.zeros(3), exx=False, mf=None, mesh=None, Gv=None,
              wrap_around=True, omega=None, **kwargs):
    '''Calculate the coarse-grained Coulomb kernel for all G-vectors, handling G=0 and exchange.
    This routine overrides get_coulG to perform interaction coarse-graining.
    '''
    exxdiv = exx
    if isinstance(exx, str):
        exxdiv = exx
    elif exx and mf is not None:
        exxdiv = mf.exxdiv

    if mesh is None:
        mesh = cell.mesh
    if 'gs' in kwargs:
        logging.warning('cell.gs is deprecated.  It is replaced by cell.mesh,'
                      'the number of PWs (=2*gs+1) along each direction.')
        mesh = [2*n+1 for n in kwargs['gs']]
    if Gv is None:
        Gv = cell.get_Gv(mesh)
    absG2 = []
    kG_0, equal2boundary_0 = compute_kG(k, Gv, wrap_around, mesh, cell)
    absG2_0 = np.einsum('gi,gi->g', kG_0, kG_0)
    for kp in lattice_kmesh:
        kG, equal2boundary = compute_kG(k + kp, Gv, wrap_around, mesh, cell)
        absG2.append(np.einsum('gi,gi->g', kG, kG))
    absG2 = np.array(absG2)

    if getattr(mf, 'kpts', None) is not None:
        kpts = mf.kpts
    else:
        kpts = k.reshape(1,3)
    Nk = len(kpts)
    fullkpts = np.zeros([kpts.shape[0]*lattice_kmesh.shape[0],3])
    for ikk, kk in enumerate(kpts):
        for ikkp, kkp in enumerate(lattice_kmesh):
            fullkpts[ikk*lattice_kmesh.shape[0] + ikkp] = kk + kkp
    Nkk = Nk * lattice_kmesh.shape[0]

    if exxdiv == 'vcut_sph':  # PRB 77 193110
        raise NotImplementedError

    elif exxdiv == 'vcut_ws':  # PRB 87, 165122
        raise NotImplementedError

    else:
        # Ewald probe charge method to get the leading term of the finite size
        # error in exchange integrals

        G0_idx = np.where(absG2_0==0)[0]
        if cell.dimension != 2 or cell.low_dim_ft_type == 'inf_vacuum':
            coulG = np.zeros(absG2.shape)
            for ikp in range(absG2.shape[0]):
                #with np.errstate(divide='ignore'):
                coulG[ikp] = 4*np.pi/absG2[ikp]
                if np.sum(np.dot(k,k)) < 1e-9: 
                    print(ikp, 4*np.pi/absG2[ikp])
                if not np.isfinite(coulG[ikp, G0_idx]) :
                    coulG[ikp, G0_idx] = 0
            coulG = np.sum(coulG, axis=0) / absG2.shape[0]
            if np.sum(np.dot(k,k)) < 1e-9: 
               print("Sum:", coulG)
            #coulG[G0_idx] = 0
            

        elif cell.dimension == 2:

            raise NotImplementedError

        elif cell.dimension == 1:
            logging.warning(cell, 'No method for PBC dimension 1, dim-type %s.'
                        '  cell.low_dim_ft_type="inf_vacuum"  should be set.',
                        cell.low_dim_ft_type)
            raise NotImplementedError


        # The divergent part of periodic summation of (ii|ii) integrals in
        # Coulomb integrals were cancelled out by electron-nucleus
        # interaction. The periodic part of (ii|ii) in exchange cannot be
        # cancelled out by Coulomb integrals. Its leading term is calculated
        # using Ewald probe charge (the function madelung below)
        if cell.dimension > 0 and exxdiv == 'ewald' and len(G0_idx) > 0:
            coulG[G0_idx] += Nkk*cell.vol*tools.madelung(cell, fullkpts)

    coulG[equal2boundary_0] = 0

    # Scale the coulG kernel for attenuated Coulomb integrals.
    # * omega is used by RangeSeparatedJKBuilder which requires ewald probe charge
    # being evaluated with regular Coulomb interaction (1/r12).
    # * cell.omega, which affects the ewald probe charge, is often set by
    # DFT-RSH functionals to build long-range HF-exchange for erf(omega*r12)/r12
    if omega is not None:
        if omega > 0:
            # long range part
            coulG *= np.exp(-.25/omega**2 * absG2)
        elif omega < 0:
            # short range part
            coulG *= (1 - np.exp(-.25/omega**2 * absG2))
    elif cell.omega > 0:
        coulG *= np.exp(-.25/cell.omega**2 * absG2)
    elif cell.omega < 0:
        raise NotImplementedError

    return coulG

def weighted_coulG_ewald(mydf, kpt, exx, mesh, omega=None):
    if omega is None:
        return df.aft.weighted_coulG(mydf, kpt, "ewald", mesh)
    return df.aft.weighted_coulG(mydf, kpt, "ewald", mesh, omega)

# a = lattice vectors / (2*pi)
@jit(nopython=True)
def kpair_reduced_lists(kptis, kptjs, kptij_idx, kmesh, a):
    nkpts = kmesh.shape[0]
    print("nkpts = ", nkpts)
    kptis = np.asarray(kptis)
    kptjs = np.asarray(kptjs)
    if kptis.shape[0] != kptjs.shape[0]:
        raise ValueError("Error: Dimension of kptis and kptjs doesn't match.")
    num_kpair = kptis.shape[0]
    conj_list = np.arange(num_kpair, dtype=np.int64)
    trans_list = np.arange(num_kpair, dtype=np.int64)
    seen = np.zeros(num_kpair, dtype=np.int64)
    for i in range(num_kpair):
        if seen[i] != 0:
            continue
        k1 = kptis[i]
        k2 = kptjs[i]
        k1_idx, k2_idx = kptij_idx[i]
        l = 0
        # conj_list: (k1, k2) = (-kk1, -kk2)
        for kk1_idx in range(k1_idx,nkpts):
            kk1 = kmesh[kk1_idx]
            kjdif = np.dot(a, k1 + kk1)
            kjdif_int = np.rint(kjdif)
            maskj = np.sum(np.abs(kjdif - kjdif_int)) < 1e-6

            if maskj == True:
                for kk2_idx in range(0,kk1_idx+1):
                    kk2 = kmesh[kk2_idx]
                    kidif = np.dot(a, k2 + kk2)
                    #kidif = np.einsum('wx,x->w', a, k2 + kk2)
                    kidif_int = np.rint(kidif)
                    maski = np.sum(np.abs(kidif - kidif_int)) < 1e-6

                    if maski == True:
                        j = int(kk1_idx * (kk1_idx + 1)/2 + kk2_idx)
                        conj_list[j] = i
                        seen[j] = 1
                        l = 1
                        break
            if l == 1:
                break
        if l == 1:
            seen[i] = 1
            continue

        # trans_list: (k1, k2) = (-kk2, -kk1)
        for kk1_idx in range(k1_idx,nkpts):
            kk1 = kmesh[kk1_idx]
            kjdif = np.dot(a, k2 + kk1)
            kjdif_int = np.rint(kjdif)
            maskj = np.sum(np.abs(kjdif - kjdif_int)) < 1e-6

            if maskj == True:
                for kk2_idx in range(0,kk1_idx+1):
                    kk2 = kmesh[kk2_idx]
                    kidif = np.dot(a, k1 + kk2)
                    kidif_int = np.rint(kidif)
                    maski = np.sum(np.abs(kidif - kidif_int)) < 1e-6

                    if maski == True:
                        j = int(kk1_idx * (kk1_idx + 1)/2 + kk2_idx)
                        trans_list[j] = i
                        seen[j] = 1
                        l = 1
                        break
            if l == 1:
                break
        seen[i] = 1

    return conj_list, trans_list


def integrals_grid(mycell, kmesh):
    a_lattice = mycell.lattice_vectors() / (2*np.pi)
    kptij_lst = [(ki, kmesh[j]) for i, ki in enumerate(kmesh) for j in range(i+1)]
    kptij_idx = [(i, j) for i in range(kmesh.shape[0]) for j in range(i+1)]
    kptij_lst = np.asarray(kptij_lst)
    kptij_idx = np.asarray(kptij_idx)
    kptis = kptij_lst[:,0]
    kptjs = kptij_lst[:,1]
    kij_conj, kij_trans = kpair_reduced_lists(kptis, kptjs, kptij_idx, kmesh, a_lattice)
    kpair_irre_list = np.argwhere(kij_conj == kij_trans)[:,0]
    num_kpair_stored = len(kpair_irre_list)
    return kptij_idx, kij_conj, kij_trans, kpair_irre_list, num_kpair_stored, kptis, kptjs

def compute_integrals(args, mycell, mydf, kmesh, nao, X_k=None, basename = "df_int", cderi_name="cderi.h5", keep=True, keep_after=False):

    kptij_idx, kij_conj, kij_trans, kpair_irre_list, num_kpair_stored, kptis, kptjs = integrals_grid(mycell, kmesh)

    mydf.kpts = kmesh
    filename = basename + "/meta.h5"
    os.system("sync") # This is needed to syncronize the NFS between nodes
    if os.path.exists(basename):
        unsafe_rm = "rm  " + basename + "/VQ*"
        os.system(unsafe_rm)
        unsafe_rm = "rm  " + basename + "/meta*"
        os.system(unsafe_rm)
        # This is needed to ensure that files are removed both on the computational and head nodes:
        os.system("sync") 
    os.system("mkdir -p " + basename) # Here "-p" is important and is needed if the if condition is trigerred
    if os.path.exists(cderi_name) and keep :
        mydf._cderi = cderi_name
    else:
        mydf._cderi_to_save = cderi_name
        mydf.build()

    auxcell = addons.make_auxmol(mycell, mydf.auxbasis)
    NQ = auxcell.nao_nr()
    print("NQ = ", NQ)

    # compute partitioning

    def compute_partitioning(tot_size, num_kpair_stored):
        # We have a guess for each fitted density upper bound of 150M
        memory = 700 if args.memory is None else args.memory
        ubound = memory * 1024 * 1024
        if tot_size > ubound :
            mult = tot_size // ubound
            chunks = num_kpair_stored // (mult+1)
            if chunks == 0 :
                print("\n\n\n Chunk size is bigger than upper memory bound per chunk you have \n\n\n")
                chunks = 1
            return chunks
        return num_kpair_stored

    single_rho_size = nao**2 * NQ * 16
    full_rho_size   = (num_kpair_stored * single_rho_size)
    chunk_size = compute_partitioning(full_rho_size, num_kpair_stored)
    print("The chunk size: ", chunk_size, " k-point pair")

    # open file to write integrals in
    if os.path.exists(filename) :
        os.remove(filename)
    data = h5py.File(filename, "w")

    # Loop over k-point pair
    # processed densities count
    cnt = 0
    # densities buffer
    buffer = np.zeros((chunk_size, NQ, nao, nao), dtype=complex)
    Lpq_mo = np.zeros((NQ, nao, nao), dtype=complex)
    chunk_indices = []
    for i in kpair_irre_list:
        k1 = kptis[i]
        k2 = kptjs[i]
        # auxiliary basis index
        s1 = 0
        for XXX in mydf.sr_loop((k1,k2), max_memory=4000, compact=False):
            LpqR = XXX[0]
            LpqI = XXX[1]
            Lpq = (LpqR + LpqI*1j).reshape(LpqR.shape[0], nao, nao)
            buffer[cnt% chunk_size, s1:s1+Lpq.shape[0], :, :] = Lpq[0:Lpq.shape[0],:,:]
            # s1 = NQ at maximum.
            s1 += Lpq.shape[0]
        cnt += 1

        # if reach chunk size: (cnt-chunk_size) equals to chunk id.
        if cnt % chunk_size == 0:
            chunk_name = basename + "/VQ_{}.h5".format(cnt - chunk_size)
            if os.path.exists(chunk_name) :
                os.remove(chunk_name)
            VQ = h5py.File(chunk_name, "w")
            VQ["{}".format(cnt - chunk_size)] = buffer.view(np.float64)
            VQ.close()
            chunk_indices.append(cnt - chunk_size)
            buffer[:] = 0.0
    # Deal the rest
    if cnt % chunk_size != 0:
        last_chunk = (num_kpair_stored // chunk_size) * chunk_size
        chunk_name = basename + "/VQ_{}.h5".format(last_chunk)
        if os.path.exists(chunk_name) :
            os.remove(chunk_name)
        VQ = h5py.File(chunk_name, "w")
        VQ["{}".format(last_chunk)] = buffer.view(np.float64)
        chunk_indices.append(last_chunk)
        VQ.close()
        buffer[:] = 0.0

    data["chunk_size"] = chunk_size
    data["chunk_indices"] = np.array(chunk_indices)
    data.attrs["__green_version__"] = imd.version("green_mbtools")
    data.close()
    if not keep_after:
        os.remove(cderi_name)
        os.system("sync")
    print("Integrals have been computed and stored into {}".format(filename))
    return kij_conj, kij_trans, kpair_irre_list, kptij_idx, num_kpair_stored

def weighted_coulG_ewald_2nd(mydf, kpt, exx, mesh):
    # this is a dirty hack
    # PySCF needs to have a full k-grid to properly compute the madelung constant
    # but we want to compute only the contribution for ki == kj to speedup this calculations
    # TODO Double check where we define full_k_mesh?
    if not hasattr(mydf.cell, 'full_k_mesh'):
         raise RuntimeError("Using wrong DF object")
    oldkpts = mydf.kpts
    mydf.kpts = mydf.cell.full_k_mesh
    coulG = df.aft.weighted_coulG(mydf, kpt, 'ewald', mesh)
    mydf.kpts = oldkpts
    return coulG

def compute_ewald_correction(args, maindf, kmesh, nao, filename = "df_ewald.h5"):
    # global full_k_mesh
    data = h5py.File(filename, "w")
    EW     = data.create_group("EW")
    EW_bar = data.create_group("EW_bar")
    # keep original method for computing Coulomb kernel
    import importlib.util as iu
    new_pyscf = iu.find_spec('pyscf.pbc.df.gdf_builder') is not None
    if new_pyscf :
        import pyscf.pbc.df.gdf_builder as gdf
        weighted_coulG_old = gdf._CCGDFBuilder.weighted_coulG
    else:
        from pyscf.pbc import df as gdf
        weighted_coulG_old = gdf.GDF.weighted_coulG

    # density-fitting w/o ewald correction for fine grid
    df2 = df.GDF(maindf.cell)
    if hasattr(df2, "_prefer_ccdf"):
        df2._prefer_ccdf = True  # Disable RS-GDF switch for new pyscf versions
    if maindf.auxbasis is not None:
        df2.auxbasis = maindf.auxbasis
    # Coulomb kernel mesh
    df2.mesh = maindf.mesh
    cderi_file_2 = ''.join(random.choices(string.ascii_uppercase + string.digits, k=10)) + ".h5"
    df2._cderi_to_save = cderi_file_2
    df2._cderi = cderi_file_2
    df2.kpts = kmesh
    df2.build()

    # densities buffer
    auxcell = addons.make_auxmol(maindf.cell, maindf.auxbasis)
    NQ = auxcell.nao_nr()
    buffer1 = np.zeros((NQ, nao, nao), dtype=np.complex128)
    buffer2 = np.zeros((NQ, nao, nao), dtype=np.complex128)
    Lpq_mo = np.zeros((NQ, nao, nao), dtype=np.complex128)
    if new_pyscf :
        gdf._CCGDFBuilder.weighted_coulG = weighted_coulG_ewald_2nd
    else:
        gdf.GDF.weighted_coulG = weighted_coulG_ewald_2nd
    # df.aft.weighted_coulG = weighted_coulG_ewald_2nd
    df1 = df.GDF(maindf.cell)
    df1.cell.full_k_mesh = maindf.kpts
    if hasattr(df1, "_prefer_ccdf"):
        df1._prefer_ccdf = True  # Disable RS-GDF switch for new pyscf versions
    if maindf.auxbasis is not None:
        df1.auxbasis = maindf.auxbasis
    # Use Ewald for divergence treatment
    df1.exxdiv = 'ewald'
    # Coulomb kernel mesh
    df1.mesh = maindf.mesh
    cderi_file_1 = ''.join(random.choices(string.ascii_uppercase + string.digits, k=10)) + ".h5"
    df1._cderi_to_save = cderi_file_1
    df1._cderi = cderi_file_1
    df1.kpts = kmesh
    df1.build()
    nk = args.nk

    # We know that G=0 contribution diverge only when q = 0
    # so we loop over (k1,k1) pairs
    for i, ki in enumerate(kmesh):
        # Change the way to compute Coulomb kernel to include G=0 correction
        # df.GDF.weighted_coulG = weighted_coulG_ewald_2nd
        if new_pyscf :
            gdf._CCGDFBuilder.weighted_coulG = weighted_coulG_ewald_2nd
        else:
            df.GDF.weighted_coulG = weighted_coulG_ewald_2nd
        s1 = 0
        # Compute three-point integrals with G=0 contribution included with Ewald correction
        for XXX in df1.sr_loop((ki,ki), max_memory=4000, compact=False):
            LpqR = XXX[0]
            LpqI = XXX[1]
            Lpq = (LpqR + LpqI*1.j).reshape(LpqR.shape[0], nao, nao)
            for G in range(Lpq.shape[0]):
                Lpq_mo[G] = Lpq[G]
            buffer1[s1:s1+Lpq.shape[0], :, :] = Lpq_mo[0:Lpq.shape[0],:,:]
            s1 += Lpq.shape[0]

        # Restore the way to compute Coulomb kernel
        # df.aft.weighted_coulG = weighted_coulG_old
        if new_pyscf:
            gdf._CCGDFBuilder.weighted_coulG = weighted_coulG_old
        else:
            df.GDF.weighted_coulG = weighted_coulG_old
        s1 = 0
        # Compute three-point integral without G=0 contribution included with Ewald correction
        # and subtract it from the computed buffer to keep pure Ewald correction only
        for XXX in df2.sr_loop((ki,ki), max_memory=4000, compact=False):
            LpqR = XXX[0]
            LpqI = XXX[1]
            Lpq = (LpqR + LpqI*1.j).reshape(LpqR.shape[0], nao, nao)
            for G in range(Lpq.shape[0]):
                Lpq_mo[G] = Lpq[G]
            buffer2[s1:s1+Lpq.shape[0], :, :] = Lpq_mo[0:Lpq.shape[0],:,:]
            s1 += Lpq.shape[0]

        # i = k1 * nk * nk + k2 * nk + k3
        k3 = int(i /(nk*nk))
        k2 = int((i % (nk*nk))/nk)
        k1 = int(i % nk)
        k_fine = (k3) * (nk)*(nk) + (k2) * (nk) + (k1)
        EW["{}".format(k_fine)] = (buffer1 - buffer2).view(np.float64)
        EW_bar["{}".format(k_fine)] = buffer2.view(np.float64)
        buffer1[:] = 0.0
        buffer2[:] = 0.0

    data.close()
    # cleanup
    # df.aft.weighted_coulG = weighted_coulG_old
    if new_pyscf:
        gdf._CCGDFBuilder.weighted_coulG = weighted_coulG_old
    else:
        gdf.GDF.weighted_coulG = weighted_coulG_old
    os.remove(cderi_file_1)
    os.remove(cderi_file_2)
    print("Ewald correction has been computed and stored into {}".format(filename))



import numpy as np
import os


def dump_input_caratheodory_data(wsample, X_iw, ifile):
    """Function to dump input data to files for caratheodory
    analytic continuation.

    Caratheodory requires input data file to be in the following format:
        iw_1 iw_2 iw_3 ... iw_n

        Real X(iw_1) Matrix

        Imag X(iw_1) Matrix

        Real X(iw_2) Matrix

        Imag X(iw_2) Matrix

        ... and so on.
    """

    nw, ns, nk, nao = X_iw.shape[:4]

    # Save dim1 for better understanding of output
    np.savetxt("dimensions.txt", np.asarray(X_iw.shape[1:3], dtype=int))

    # Dump input data to file
    dim = 0
    for js in range(ns):
        for jk in range(nk):
            # make the directory for given (js, jk) point
            if not os.path.exists(str(dim)):
                os.mkdir(str(dim))
            # Write the input data to file
            output_fname = str(dim) + '/' + ifile
            with open(output_fname, 'w') as fs:
                # write frequencies in the first line
                for w_j in wsample:
                    fs.write(str(w_j) + '\t')
                fs.write('\n\n')
                for jw in range(nw):
                    # Real part for jw-th frequency
                    for p in range(nao):
                        for q in range(nao):
                            fs.write(str(X_iw[jw, js, jk, p, q].real) + '\t')
                        fs.write('\n')
                    fs.write('\n')
                    # Imag part for jw-th frequency
                    for p in range(nao):
                        for q in range(nao):
                            fs.write(str(X_iw[jw, js, jk, p, q].imag) + '\t')
                        fs.write('\n')
                    fs.write('\n')
            dim += 1

    return None


def load_caratheodory_data(matrix_file, spectral_file, X_dims):
    """Load output data from caratheodory analytic continuation.
    The spectral function output file has the format:
        w1 XA(w1 + i eta)
        w2 XA(w2 + i eta)
        ... and so on.
    And the complex matrix output file has the format:
        w1 Re.Xc[11](w1 + i eta) Im.Xc[11](w1 + i eta) Re.Xc[12](w1 + i eta)
        w2 Re.Xc[11](w2 + i eta) Im.Xc[11](w2 + i eta) Re.Xc[12](w2 + i eta)
        ... and so on.
    """

    # Dimensions
    _, ns, nk, nao = X_dims[:4]
    dim1 = ns * nk

    # Load the spectral function data
    dump_A = False
    try:
        XA_w = np.loadtxt("0/{}".format(spectral_file))
        freqs = XA_w[:, 0]
    except IOError:
        pass
    else:
        dump_A = True

    if dump_A:
        XA_w = np.zeros((freqs.shape[0], dim1))
        for d1 in range(dim1):
            # Read X_w data
            try:
                X_wsk = np.loadtxt("{}/{}".format(d1, spectral_file))
                XA_w[:, d1] = X_wsk[:, 1]
            except IOError:
                print(
                    "{} is missing in {} folder. Analytical continuation \
                    may have failed at that point.".format(spectral_file, d1)
                )
        # reshape the spectral data
        XA_w = XA_w.reshape((freqs.shape[0],) + (ns, nk))
    else:
        print("All AC fails. Will not dump to DOS.h5")

    # Load the complex matrix data
    dump_c = False
    try:
        Xc_w = np.loadtxt("0/{}".format(matrix_file))
    except IOError:
        pass
    else:
        dump_c = True

    if dump_c:
        Xc_w = np.zeros((freqs.shape[0], dim1, nao, nao), dtype=complex)
        for d1 in range(dim1):
            # Read X_c data
            try:
                X_wsk = np.loadtxt("{}/{}".format(d1, matrix_file))
                for jw in range(len(freqs)):
                    # need to convert the real + imag data into complex type
                    Xc_imtd = X_wsk[jw, 1:].view(complex)
                    Xc_w[jw, d1, :, :] = Xc_imtd.reshape((nao, nao))
            except IOError:
                print(
                    "{} is missing in {} folder. Analytical continuation \
                    may have failed at that point.".format(spectral_file, d1)
                )
        # reshape the complex matrix data
        Xc_w = Xc_w.reshape((freqs.shape[0],) + (ns, nk, nao, nao))

    return freqs, Xc_w, XA_w



import argparse
import logging
import os

import h5py
import numpy as np
import pyscf.lib.chkfile as chk
from io import StringIO
from numba import jit
from pyscf import gto as mgto
from pyscf import scf as mscf
from pyscf import dft as mdft
from pyscf import df as mdf
from pyscf.pbc import tools, gto, df, scf, dft

import importlib.metadata as imd

from . import integral_utils as int_utils

def construct_rmesh(nkx, nky, nkz):
    '''
    Construct real space grid centered around (0,0,0)
    '''
    Lx, Ly, Lz = (nkx-1)//2, (nky-1)//2, (nkz-1)//2 # nk=6, L=2
    leftx, lefty, leftz = (nkx-1)%2, (nky-1)%2, (nkz-1)%2 # left = 1

    rx = np.linspace(-Lx, Lx+leftx, nkx, endpoint=True) # -2,-1,0,1,2,3
    ry = np.linspace(-Ly, Ly+lefty, nky, endpoint=True)
    rz = np.linspace(-Lz, Lz+leftz, nkz, endpoint=True)
    RX, RY, RZ = np.meshgrid(rx, ry, rz)
    rmesh = np.array([RX.flatten(), RY.flatten(), RZ.flatten()]).T

    return rmesh

def extract_ase_data(a, atoms):
    '''
    For a given data in XYZ format generate parameters in ASE format
    
    :param a: string containing lattice vectros in XYZ format
    :param atoms: string containin atoms positions in XYZ format
    '''
    symbols = []
    positions = []
    lattice_vectors = np.genfromtxt(StringIO(a))
    for atom in atoms.splitlines():
        if len(atom) == 0 or len(atom.split()) != 4:
            continue
        atom = atom.strip()
        symbol = atom.split()[0]
        position = atom.split(symbol)[1].strip()
        symbols.append(symbol)
        position = [float(p) for p in position.split()]
        positions.append(np.dot(np.linalg.inv(lattice_vectors), position).tolist())
    return (lattice_vectors, symbols, positions)

def print_high_symmetry_points(cell, args):
    '''
    For a given unit cell and simulation parameters generate and print list of lattice special points
    '''
    import ase.spacegroup
    lattice_vectors, symbols, positions = extract_ase_data(args.a, args.atom)
    cc = ase.spacegroup.crystal(symbols, positions, cellpar=ase.geometry.cell_to_cellpar(lattice_vectors))
    space_group = ase.spacegroup.get_spacegroup(cc)
    lat = cc.cell.get_bravais_lattice()
    special_points = lat.get_special_points()
    print("List of special points: {}".format(special_points))

def check_high_symmetry_path(cell, args):
    '''
    Check that selected high-symmetry path is correct for the chosen simulation parameters

    :param cell: unit cell object
    :param args: simulation parameters
    '''
    if args.high_symmetry_path is None:
        return
    import ase.spacegroup
    lattice_vectors, symbols, positions = extract_ase_data(args.a, args.atom)
    logging.info(f"parse: {lattice_vectors} {symbols} {positions}")
    cc = ase.spacegroup.crystal(symbols, positions, cellpar=ase.geometry.cell_to_cellpar(lattice_vectors))
    space_group = ase.spacegroup.get_spacegroup(cc)
    lat = cc.cell.get_bravais_lattice()
    special_points = lat.get_special_points()
    path = args.high_symmetry_path
    for sp in special_points.keys():
        path = path.replace(sp, "")
    path = path.replace(",", "")
    if path != "":
        raise RuntimeError(("Chosen high symmetry path {} has invalid special points {}. Valid "
                            "special points are {} ").format(args.high_symmetry_path, path, special_points.keys()))

def high_symmetry_path(cell, args):
    '''
    Compute high-symmetry k-path

    :param cell: unit-cell object
    :param args: simulation parameters
    :return: Points on the chosen high-symmetry path and corresponding non-interacting Hamiltonian and overlap matrix
    '''
    if args.high_symmetry_path is None:
        return [None, None, None]
    import ase
    lattice_vectors, symbols, positions = extract_ase_data(args.a, args.atom)
    path = args.high_symmetry_path
    kpath = ase.dft.kpoints.bandpath(args.high_symmetry_path, lattice_vectors, npoints=args.high_symmetry_path_points)
    kmesh = cell.get_abs_kpts(kpath.kpts)
    new_mf    = dft.KUKS(cell,kmesh).density_fit()
    H0_hs = new_mf.get_hcore()
    Sk_hs = new_mf.get_ovlp()
    return [kmesh, H0_hs, Sk_hs]

def transform(Z, X, X_inv):
    '''
    Transform Z into X basis
    :param Z: Object to be transformed
    :param X: Transformation matrix
    :param X_inv: Inverse transformation matrix
    :return: Z in new basis
    '''
    Z_X = np.zeros(Z.shape, dtype=np.complex128)
    maxdiff = -1
    for ss in range(Z.shape[0]):
        for ik in range(Z.shape[1]):
            Z_X[ss,ik] = np.einsum('ij,jk...,kl->il...', X[ik], Z[ss, ik], X[ik].T.conj())

            Z_restore = np.dot(X_inv[ik], np.dot(Z_X[ss, ik], X_inv[ik].conj().T))
            diff = np.max(np.abs(Z[ss, ik] - Z_restore))
            maxdiff = max(maxdiff, diff)

            if not np.allclose(Z[ss, ik], Z_restore, atol=1e-12, rtol=1e-12) :
                error = "Orthogonal transformation failed. Max difference between origin and restored quantity is {}".format(np.max(np.abs(Z[ss,ik] - Z_restore)))
                raise RuntimeError(error)
    logging.info(f"Maximum difference between Z and Z_restore {maxdiff}")
    return Z_X

def fold_back_to_1stBZ(kpts):
    '''
    Map each k-point from a given list of scaled k-points into the first Brillouin zone

    :param kpts: list of k-points
    '''
    nkpts = len(kpts)
    for i, ik in enumerate(kpts):
        kpts[i] = np.array([wrap_1stBZ(kk) for kk in ik])
    return kpts

def inversion_sym(kmesh_scaled):
    '''
    For a given list of the scaled k-point in the full Brillouin zone select k-points that are 
    equivalent by the time-reversal symmetry and return them and their corresponding weight and index
    '''
    ind = np.arange(np.shape(kmesh_scaled)[0])
    weight = np.zeros(np.shape(kmesh_scaled)[0])
    for i, ki in enumerate(kmesh_scaled):
        ki = [wrap_1stBZ(l) for l in ki]
        kmesh_scaled[i] = ki

    # Time-reversal symmetry
    Inv = (-1) * np.identity(3)
    for i, ki in enumerate(kmesh_scaled):
        ki = np.dot(Inv,ki)
        ki = [wrap_1stBZ(l) for l in ki]
        for l, kl in enumerate(kmesh_scaled[:i]):
            if np.allclose(ki,kl):
                ind[i] = l
                break

    uniq = np.unique(ind, return_counts=True)
    for i, k in enumerate(uniq[0]):
        weight[k] = uniq[1][i]
    ir_list = uniq[0]

    # Mark down time-reversal-reduced k-points
    conj_list = np.zeros(len(kmesh_scaled))
    for i, k in enumerate(ind):
        if i != k:
            conj_list[i] = 1

    return ir_list, ind, weight, conj_list

def wrap_k(k):
    while k < 0 :
        k = 1 + k
    while (k - 9.9999999999e-1) > 0.0 :
        k = k - 1
    return k

def parse_basis(basis_list):
    '''
    Parse information about chosen basis sets
    '''
    logging.debug(f"{basis_list}, {len(basis_list) % 2}")
    if len(basis_list) % 2 == 0:
        b = {}
        for atom_i in range(0, len(basis_list), 2):
            bas_i = basis_list[atom_i + 1]
            if os.path.exists(bas_i) :
                with open(bas_i) as bfile:
                    bas = mgto.parse(bfile.read())
            # if basis specified as a standard basis
            else:
                bas = bas_i
            b[basis_list[atom_i]] = bas
        return b
    else:
        return basis_list[0]

def parse_geometry(g):
    '''
    parse geometry of the system
    '''
    res = ""
    if os.path.exists(g) :
        with open(g) as gf:
            res = gf.read()
    else:
        res = g
    return res

def save_data(args, mycell, mf, kmesh, ind, weight, num_ik, ir_list, conj_list, Nk, nk, NQ, F, S, T, hf_dm,
              Zs, last_ao):
    '''
    Save data in Green/WeakCoupling format into a hdf5 file
    '''
    kptij_idx, kij_conj, kij_trans, kpair_irre_list, num_kpair_stored, kptis, kptjs = int_utils.integrals_grid(mycell, kmesh)
    logging.info(f"number of reduced k-pairs: {num_kpair_stored}")
    inp_data = h5py.File(args.output_path, "w")
    inp_data["grid/k_mesh"] = kmesh
    inp_data["grid/k_mesh_scaled"] = mycell.get_scaled_kpts(kmesh)
    inp_data["grid/index"] = ind
    inp_data["grid/weight"] = weight
    inp_data["grid/ink"] = num_ik
    inp_data["grid/nk"] = nk
    inp_data["grid/ir_list"] = ir_list
    inp_data["grid/conj_list"] = conj_list
    inp_data["grid/conj_pairs_list"] = kij_conj
    inp_data["grid/trans_pairs_list"] = kij_trans
    inp_data["grid/kpair_irre_list"] = kpair_irre_list
    inp_data["grid/kpair_idx"] = kptij_idx
    inp_data["grid/num_kpair_stored"] = num_kpair_stored
    inp_data["HF/Nk"] = Nk
    inp_data["HF/nk"] = nk
    inp_data["HF/Energy"] = mf.e_tot
    inp_data["HF/Energy_nuc"] = mf.energy_nuc()
    inp_data["HF/Fock-k"] = F.view(np.float64).reshape(F.shape[0], F.shape[1], F.shape[2], F.shape[3], 2)
    inp_data["HF/Fock-k"].attrs["__complex__"] = np.int8(1)
    inp_data["HF/S-k"] = S.view(np.float64).reshape(S.shape[0], S.shape[1], S.shape[2], S.shape[3], 2)
    inp_data["HF/S-k"].attrs["__complex__"] = np.int8(1)
    inp_data["HF/H-k"] = T.view(np.float64).reshape(T.shape[0], T.shape[1], T.shape[2], T.shape[3], 2)
    inp_data["HF/H-k"].attrs["__complex__"] = np.int8(1)
    inp_data["HF/madelung"] = tools.pbc.madelung(mycell, kmesh)
    inp_data["HF/mo_energy"] = mf.mo_energy
    inp_data["HF/mo_coeff"] = mf.mo_coeff
    inp_data["mulliken/Zs"] = Zs
    inp_data["mulliken/last_ao"] = last_ao
    inp_data["params/nao"] = S.shape[2]
    inp_data["params/nso"] = S.shape[2]
    inp_data["params/ns"] = S.shape[0]
    inp_data["params/nel_cell"] = mycell.nelectron
    inp_data["params/nk"] = kmesh.shape[0]
    inp_data["params/NQ"] = NQ
    inp_data.attrs["__green_version__"] = imd.version("green_mbtools")
    inp_data.close()
    chk.save(args.output_path, "Cell", mycell.dumps())
    inp_data = h5py.File("dm.h5", "w")
    inp_data["HF/dm-k"] = hf_dm.view(np.float64).reshape(hf_dm.shape[0], hf_dm.shape[1], hf_dm.shape[2], hf_dm.shape[3], 2)
    inp_data["HF/dm-k"].attrs["__complex__"] = np.int8(1)
    inp_data["dm_gamma"] = hf_dm[:, 0, :, :]
    inp_data.close()


def orthogonalize(mydf, orth, X_k, X_inv_k, F, T, hf_dm, S):
    '''
    Transform Fock-matrix, non-interacting Hamiltonian, density matrix and overlap matrix into an orthogonal basis
    '''
    maxdiff = -1
    old_shape = [-1, -1]
    for ik, k in enumerate(mydf.kpts):
        if orth == 0:
            X_inv_k.append(np.eye(F.shape[2], dtype=np.complex128))
            X_k.append(np.eye(F.shape[2], dtype=np.complex128))
            continue

        s_ev, s_eb = np.linalg.eigh(S[0, ik])

        # Remove all eigenvalues < threshold
        istart = s_ev.searchsorted(1e-9)
        s_sqrtev = np.sqrt(s_ev[istart:])

        x_pinv = s_eb[:, istart:] * s_sqrtev
        x = (s_eb[:, istart:].conj() * 1 / s_sqrtev).T
        n_ortho, n_nonortho = x.shape
        if old_shape[0] >= 0 and n_ortho != old_shape[0] and n_nonortho != old_shape[1]:
            raise RuntimeError("Error!!! Different k-point have different number of orthogonal basis.")

        old_shape[0] = n_ortho
        old_shape[1] = n_nonortho

        X_inv_k.append(x_pinv.copy())
        X_k.append(x.copy())

        diff = np.eye(n_nonortho) - np.dot(x, x_pinv)
        diff_max = np.max(np.abs(diff))
        maxdiff = max(maxdiff, diff_max)
    logging.info(f"max diff from identity {maxdiff}")
    X_inv_k = np.asarray(X_inv_k).reshape(F.shape[1:])
    X_k = np.asarray(X_k).reshape(F.shape[1:])
    # Orthogonalization
    if orth == 1:
        F = transform(F, X_k, X_inv_k)
        # S     = transform(S, X_k, X_inv_k)
        T = transform(T, X_k, X_inv_k)
        hf_dm = transform(hf_dm, X_inv_k, X_k)

        S = np.array([np.eye(F.shape[-1], dtype=np.complex128)] * F.shape[1])
        S = np.array([S, S])

    return X_k, X_inv_k, S, F, T, hf_dm

def add_common_params(parser):
    '''
    Define common command line arguments for Green python module
    '''
    parser.add_argument("--atom", type=parse_geometry, help="poistions of atoms", required=True)
    parser.add_argument("--Nk", type=int, default=0, help="number of plane-waves in each direction for integral evaluation")
    parser.add_argument("--basis", type=str, nargs="*", help="basis sets definition. First specify atom then basis for this atom", required=True)
    parser.add_argument("--auxbasis", type=str, nargs="*", default=[None], help="auxiliary basis")
    parser.add_argument("--ecp", type=str, nargs="*", default=[None], help="effective core potentials")
    parser.add_argument("--xc", type=str, nargs="*", default=[None], help="XC functional")
    parser.add_argument("--dm0", type=str, default=None, help="initial guess for density matrix")
    parser.add_argument("--df_int", type=int, default=1, help="prepare density fitting integrals or not")
    parser.add_argument("--int_path", type=str, default="df_int", help="path to store ewald corrected integrals")
    parser.add_argument("--hf_int_path", type=str, default="df_hf_int", help="path to store hf integrals")
    parser.add_argument("--output_path", type=str, default="input.h5", help="output file with initial data")
    parser.add_argument("--orth", type=int, default=0, help="Transform to orthogonal basis or not. 0 - no orthogonal transformation, 1 - data is in orthogonal basis.")
    parser.add_argument("--beta", type=float, default=None, help="Emperical parameter for even-Gaussian auxiliary basis")
    parser.add_argument("--active_space", type=int, nargs='+', default=None, help="active space orbitals")
    parser.add_argument("--spin", type=int, default=0, help="Local spin")
    parser.add_argument("--restricted", type=lambda x: (str(x).lower() in ['true','1', 'yes']), default='false', help="Spin restricted calculations.")
    parser.add_argument("--print_high_symmetry_points", default=False, action='store_true', help="Print available high symmetry points for current system and exit.")
    parser.add_argument("--high_symmetry_path", type=str, default=None, help="High symmetry path")
    parser.add_argument("--high_symmetry_path_points", type=int, default=0, help="Number of points for high symmetry path")
    parser.add_argument("--memory", type=int, default=700, help="Memory bound for integral chunk in MB")
    parser.add_argument("--grid_only", type=lambda x: (str(x).lower() in ['true','1', 'yes']), default='false', help="Only recompute k-grid points")
    parser.add_argument("--diffuse_cutoff", type=float, default=0.0, help="Remove the diffused Gaussians whose exponents are less than the cutoff")
    parser.add_argument("--damping", type=float, default=0.0, help="Damping factor for mean-field iterations")
    parser.add_argument("--max_iter", type=int, default=100, help="Maximum number of iterations in the SCF loop")
    parser.add_argument("--keep_cderi", type=lambda x: (str(x).lower() in ['true','1', 'yes']), default='false', help="Keep generated cderi files.")
    parser.add_argument("--job", choices=["init", "sym_path", "ewald_corr"], default="init", nargs="+")
    parser.add_argument("--finite_size_kind", choices=["ewald", "gf2", "gw", "gw_s", "coarse_grained"], default="ewald", nargs="+", 
                              help="Two body finite-size correction. Be default computes the second set of integrals that include simple ewald correction.")
    parser.add_argument("--x2c", type=lambda x: (str(x).lower() in ['true','1', 'yes']), default='false', help="enable X2C calculations")

def add_pbc_params(parser):
    '''
    Define PBC-specific command line arguments for Green python module
    '''
    parser.add_argument("--a", type=parse_geometry, help="lattice geometry", required=True)
    parser.add_argument("--nk", type=int, help="number of k-points in each direction", required=True)
    parser.add_argument("--pseudo", type=str, nargs="*", default=[None], help="pseudopotential")
    parser.add_argument("--shift", type=float, nargs=3, default=[0.0, 0.0, 0.0], help="mesh shift")
    parser.add_argument("--center", type=float, nargs=3, default=[0.0, 0.0, 0.0], help="mesh center")
    parser.add_argument("--symm", type=lambda x: (str(x).lower() in ['true','1', 'yes']), default='true', help="Use inversion symmetry")

def init_mol_params():
    '''
    Initialize argparse.ArgumentParser for Green/WeakCoupling python module and return a prased parameters map with parameters specific for molecular calculations
    '''
    parser = argparse.ArgumentParser(description="Green/WeakCoupling initialization script")
    add_common_params(parser)
    args = parser.parse_args()
    args.basis = parse_basis(args.basis)
    args.auxbasis = parse_basis(args.auxbasis)
    args.ecp = parse_basis(args.ecp)
    args.xc = parse_basis(args.xc)
    if args.x2c and args.restricted:
        raise RuntimeError("X2C calculation can not be spin restricted")
    if args.xc is not None:
        if args.x2c :
            args.mean_field = mdft.GKS
        else:
            args.mean_field = mdft.RKS if args.restricted else mdft.UKS
    else:
        if args.x2c:
            args.mean_field = mscf.GHF
        else :
            args.mean_field = mscf.RHF if args.restricted else mscf.UHF
    args.ns = 1 if args.restricted else 2
    # parameters needed to create empty grid
    args.a = [[1,0,0],[0,1,0],[0,0,1]]
    args.nk = 1
    args.shift =  [0.,0.,0.]
    args.center = [0.,0.,0.]
    return args

def init_pbc_params():
    '''
    Initialize argparse.ArgumentParser for Green/WeakCoupling python module and return a prased parameters map with parameters specific for periodic calculations
    '''
    parser = argparse.ArgumentParser(description="Green/WeakCoupling initialization script")
    add_common_params(parser)
    add_pbc_params(parser)
    args = parser.parse_args()
    args.basis = parse_basis(args.basis)
    args.auxbasis = parse_basis(args.auxbasis)
    args.ecp = parse_basis(args.ecp)
    args.pseudo = parse_basis(args.pseudo)
    args.xc = parse_basis(args.xc)
    if args.x2c and args.restricted:
        raise RuntimeError("X2C calculation can not be spin restricted")
    if args.xc is not None:
        if args.x2c :
            args.mean_field = dft.KGKS
        else:
            args.mean_field = dft.KRKS if args.restricted else dft.KUKS
    else:
        if args.x2c:
            args.mean_field = scf.KGHF
        else :
            args.mean_field = scf.KRHF if args.restricted else scf.KUHF
    args.ns = 1 if args.restricted or args.x2c else 2
    return args

def mol_cell(args):
    '''
    Initialize PySCF unit cell object for a given system
    '''
    c = mgto.M(
        atom = args.atom,
        unit = 'A',
        basis = args.basis,
        ecp = args.ecp,
        verbose = 7,
        spin = args.spin
    )
    return c

def pbc_cell(args):
    '''
    Initialize PySCF unit cell object for a given system
    '''
    c = gto.M(
        a = args.a,
        atom = args.atom,
        unit = 'A',
        basis = args.basis,
        ecp = args.ecp,
        pseudo = args.pseudo,
        verbose = 7,
        spin = args.spin,
#        exp_to_discard = args.diffuse_cutoff
    )
    _a = c.lattice_vectors()
    c.exp_to_discard = args.diffuse_cutoff
    c.build()
    if np.linalg.det(_a) < 0:
        raise "Lattice are not in right-handed coordinate system. Please correct your lattice vectors"
    return c

def wrap_1stBZ(k):
    '''
    wrap scaled k into [-0.5,0.5) range
    :param k: value of k-point at some dimension
    '''
    while k < -0.5 :
        k = k + 1
    while (k - 4.9999999999e-1) > 0.0 :
        k = k - 1
    return k

def init_k_mesh(args, mycell):
    '''
    init k-points mesh for GDF

    :param args: script arguments
    :param mycell: unit cell for simulation
    :return: kmesh, k_ibz, ir_list, conj_list, weight, ind, num_ik
    '''
    if args.center is None:
       args.center = [0,0,0]
    if args.shift is None:
       args.shift = [0,0,0]
    kmesh = mycell.make_kpts([args.nk, args.nk, args.nk], scaled_center=args.center)
    for i, kk in enumerate(kmesh):
        ki = kmesh[i]
        ki = mycell.get_scaled_kpts(ki) + args.shift
        ki = [wrap_k(l) for l in ki]
        kmesh[i] = mycell.get_abs_kpts(ki)
    for i, ki in enumerate(kmesh):
        ki = mycell.get_scaled_kpts(ki)
        ki = [wrap_k(l) for l in ki]
        ki = mycell.get_abs_kpts(ki)
        kmesh[i] = ki

    logging.debug(kmesh)
    logging.info(mycell.get_scaled_kpts(kmesh))

    if not args.symm :
        nkpts = kmesh.shape[0]
        weight = np.ones(nkpts)
        ir_list = np.array(range(nkpts))
        ind = np.array(range(nkpts))
        conj_list = np.zeros(nkpts)
        k_ibz = np.copy(kmesh)
        num_ik = nkpts
        return kmesh, k_ibz, ir_list, conj_list, weight, ind, num_ik

    logging.info("Compute irreducible k-points")


    k_ibz = mycell.make_kpts([args.nk,args.nk,args.nk], scaled_center=args.center)
    ind = np.arange(np.shape(k_ibz)[0])
    weight = np.zeros(np.shape(k_ibz)[0])
    for i, ki in enumerate(k_ibz):
        ki = mycell.get_scaled_kpts(ki)
        ki = [wrap_1stBZ(l) for l in ki]
        k_ibz[i] = ki

    # Time-reversal symmetry
    Inv = (-1) * np.identity(3)
    for i, ki in enumerate(k_ibz):
        ki = np.dot(Inv,ki)
        ki = [wrap_1stBZ(l) for l in ki]
        for l, kl in enumerate(k_ibz[:i]):
            if np.allclose(ki,kl):
                k_ibz[i] = kl
                ind[i] = l
                break

    uniq = np.unique(ind, return_counts=True)
    for i, k in enumerate(uniq[0]):
        weight[k] = uniq[1][i]
    ir_list = uniq[0]

    # Mark down time-reversal-reduced k-points
    conj_list = np.zeros(args.nk**3)
    for i, k in enumerate(ind):
        if i != k:
            conj_list[i] = 1
    num_ik = np.shape(uniq[0])[0]

    return kmesh, k_ibz, ir_list, conj_list, weight, ind, num_ik



def read_dm(dm0, dm_file):
    '''
    Read density matrix from smaller kmesh
    '''
    nao  = dm0.shape[-1]
    nkpts = dm0.shape[1]
    dm   = np.zeros((2,nao,nao),dtype=np.complex128)
    f    = h5py.File(dm_file, 'r')
    dm[:,:,:] = f['/dm_gamma'][:]
    f.close()
    dm_kpts = np.repeat(dm[:,None, :, :], nkpts, axis=1)
    return dm_kpts

def solve_mean_field(args, mydf, mycell):
    '''
    Obtain pySCF mean-field solution for a given parameters, unit-cell object and density-fitting object
    '''
    logging.info("Solve Mean-field")
    # prepare and solve DFT
    mf    =  args.mean_field(mycell,mydf.kpts).density_fit().x2c1e() if args.x2c  else args.mean_field(mycell,mydf.kpts).density_fit() # if args.xc is not None else scf.KUHF(mycell,mydf.kpts).density_fit()
    if args.xc is not None:
        mf.xc = args.xc
    #mf.max_memory = 10000
    mydf._cderi = "cderi.h5"
    mf.kpts = mydf.kpts
    mf.with_df = mydf
    mf.diis_space = 16
    mf.damp = args.damping
    mf.max_cycle = args.max_iter
    mf.chkfile = 'tmp.chk'
    if os.path.exists("tmp.chk"):
        init_dm = mf.from_chk('tmp.chk')
        mf.kernel(init_dm)
    elif args.dm0 is not None:
        init_dm = mf.get_init_guess()
        init_dm = read_dm(init_dm, args.dm0)
        mf.kernel(init_dm)
    else:
        mf.kernel()
    mf.analyze()
    return mf

def solve_mol_mean_field(args, mydf, mycell):
    '''
    Obtain pySCF mean-field solution for a given parameters, unit-cell object and density-fitting object
    '''
    logging.info("Solve LDA")
    # prepare and solve DFT
    
    mf    = args.mean_field(mycell).x2c1e() if args.x2c else  args.mean_field(mycell).density_fit() # if args.xc is not None else scf.KUHF(mycell,mydf.kpts).density_fit()
    if args.xc is not None:
        mf.xc = args.xc
    #mf.max_memory = 10000
    #mydf._cderi = "cderi.h5"
    mf.with_df._cderi_to_save = "cderi_mol.h5"
    mf.diis_space = 16
    mf.damp = args.damping
    mf.max_cycle = args.max_iter
    mf.chkfile = 'tmp.chk'
    if os.path.exists("tmp.chk"):
        init_dm = mf.from_chk('tmp.chk')
        mf.kernel(init_dm)
    elif args.dm0 is not None:
        init_dm = mf.get_init_guess()
        init_dm = read_dm(init_dm, args.dm0)
        mf.kernel(init_dm)
    else:
        mf.kernel()
    mf.analyze()
    return mf


def store_k_grid(args, mycell, kmesh, k_ibz, ir_list, conj_list, weight, ind, num_ik):
    '''
    Store reciprocal space information into a hdf5 file in Green'WeakCoupling format
    '''
    inp_data = h5py.File(args.output_path, "a")
    nk = kmesh.shape[0]
    ink = k_ibz.shape[0]
    kptij_idx, kij_conj, kij_trans, kpair_irre_list, num_kpair_stored, kptis, kptjs = int_utils.integrals_grid(mycell, kmesh)
    logging.info(f"number of reduced k-pairs: {num_kpair_stored}")
    if not "grid" in inp_data:
        inp_data.create_group("grid")
    grid_grp = inp_data["grid"]
    data = [kmesh, mycell.get_scaled_kpts(kmesh), ind, weight, num_ik, nk, ir_list, conj_list,
               kij_conj, kij_trans, kpair_irre_list, kptij_idx, num_kpair_stored]
    names = ["k_mesh", "k_mesh_scaled", "index", "weight", "ink", "nk", "ir_list", "conj_list",
                "conj_pairs_list", "trans_pairs_list", "kpair_irre_list", "kpair_idx", "num_kpair_stored" ]
    for i, name in enumerate(names):
        if name in grid_grp:
            grid_grp[name][...] = data[i]
        else:
            grid_grp[name] = data[i]
    inp_data.close()

def construct_mol_gdf(args, mycell):
    '''
    Construct Gaussian Density Fitting obejct for a given parameters and unit cell.
    We make sure to disable range-separeting implementation
    '''
    # Use gaussian density fitting to get fitted densities
    mydf = df.GDF(mycell, mycell.kpts)
    if args.auxbasis is not None:
        mydf.auxbasis = args.auxbasis
    elif args.beta is not None:
        mydf.auxbasis = df.aug_etb(mycell, beta=args.beta)
    # Coulomb kernel mesh
    if args.Nk > 0:
        mydf.mesh = [args.Nk, args.Nk, args.Nk]
    return mydf


def construct_gdf(args, mycell, kmesh=None):
    '''
    Construct Gaussian Density Fitting obejct for a given parameters and unit cell.
    We make sure to disable range-separeting implementation
    '''
    # Use gaussian density fitting to get fitted densities
    mydf = df.GDF(mycell)
    if hasattr(mydf, "_prefer_ccdf"):
        mydf._prefer_ccdf = True  # Disable RS-GDF switch for new pyscf versions 
    if args.auxbasis is not None:
        mydf.auxbasis = args.auxbasis
    elif args.beta is not None:
        mydf.auxbasis = df.aug_etb(mycell, beta=args.beta)
    # Coulomb kernel mesh
    if args.Nk > 0:
        mydf.mesh = [args.Nk, args.Nk, args.Nk]
    if kmesh is not None:
        mydf.kpts = kmesh
    return mydf


def compute_ewald_correction(args, cell, kmesh, filename):
    # Use gaussian density fitting to get fitted densities
    mydf = df.GDF(cell)
    if args.auxbasis is not None:
        mydf.auxbasis = args.auxbasis
    elif args.beta is not None:
        mydf.auxbasis = df.aug_etb(cell, beta=args.beta)
    # Coulomb kernel mesh
    if args.Nk > 0:
        mydf.mesh = [args.Nk, args.Nk, args.Nk]
    int_utils.compute_ewald_correction(args, mydf, kmesh, cell.nao_nr(), filename)

def compute_df_int_dca(args, mycell, kmesh, lattice_kmesh, nao, X_k):
    '''
    Generate density-fitting integrals for correlated methods using q-averaging over the super-lattice points to compensate finite-size error
    
    :param args: simulation parameters
    :param mycell: unit cell object
    :param kmesh: reciprocal space grid
    :param lattice_kmesh: super-lattice k-points
    :param nao: number of atomic orbitals in the unit cell
    :param X_k: trasformation matrix for projection onto an orthogonal space
    '''

    if not bool(args.df_int):
        return
    mydf = construct_gdf(args, mycell, kmesh)
    # Use Ewald for divergence treatment
    mydf.exxdiv = 'ewald'
    weighted_coulG_old = df.GDF.weighted_coulG
    df.GDF.weighted_coulG = int_utils.weighted_coulG_ewald
    old_get_coulG = tools.get_coulG
    tools.get_coulG = lambda cell, k=np.zeros(3), exx=False, mf=None, mesh=None, Gv=None, wrap_around=True, omega=None, **kwargs: int_utils.get_coarsegrained_coulG(lattice_kmesh, cell, k, exx, mf, mesh, Gv,
              wrap_around, omega, **kwargs)

    kij_conj, kij_trans, kpair_irre_list, kptij_idx, num_kpair_stored = int_utils.compute_integrals(mycell, mydf, kmesh, nao, X_k, args.int_path, "cderi_ewald_dca.h5", False)

    mydf = None
    mydf = construct_gdf(args, mycell, kmesh)
    df.GDF.weighted_coulG = weighted_coulG_old
    int_utils.compute_integrals(mycell, mydf, kmesh, nao, X_k, args.hf_int_path, "cderi_dca.h5", False)

    tools.get_coulG = old_get_coulG




from setuptools import setup, Extension
import sys
import platform

include_dirs    = ['/usr/local/include', '/opt/homebrew/include', '/usr/local/include/eigen3', '/opt/homebrew/include/eigen3' ] if sys.platform=='darwin' else ['/usr/include/eigen3']
library_dirs    = ['/opt/homebrew/lib','/usr/local/lib' ] if sys.platform=='darwin' else []
compile_flags   = ["-std=c++17"]
extra_link_args = []
if sys.platform=='darwin' and int(platform.mac_ver()[0].split('.')[0]) >= 14 :
    compile_flags+=['-arch', 'x86_64', '-arch', 'arm64']
    extra_link_args+=["-undefined", "dynamic_lookup"]

# Build Nevanlinna extension
nevanlinna = Extension(
    'green_mbtools.pesto.nevanlinna',
    sources=['Nevanlinna/nevanlinna.cpp' ],
    include_dirs=include_dirs + ['Nevanlinna',],
    library_dirs=library_dirs,
    depends=['Nevanlinna/schur.h','Nevanlinna/nevanlinna.h','Nevanlinna/gmp_float.h'],
    libraries=['gmp', 'gmpxx'],
    extra_compile_args=compile_flags,
    extra_link_args=extra_link_args
)

# Build Caratheodory extension
caratheodory = Extension(
    'green_mbtools.pesto.caratheodory',
    sources=['Caratheodory/caratheodory.cpp'],
    include_dirs=include_dirs + ['Caratheodory',],
    library_dirs=library_dirs,
    depends=['Caratheodory/carath.h','Caratheodory/iter.h','Caratheodory/mpfr_float.h'],
    libraries=['gmp', 'gmpxx', 'mpfr'],
    extra_compile_args=compile_flags,
    extra_link_args=extra_link_args
)

setup(
   ext_modules=[nevanlinna, caratheodory]
)



import numpy as np

from . import ft
from . import dyson


# Only work for full_bz object
def interpolate(obj_k, kmesh, kpts_inter, dim=3, hermi=False, debug=False):
    """Interpolate obj_k[ns, nk, nao, nao] from kmesh to kpts_inter
    using Wannier interpolation.

    NOTE: all the k-points are in scaled units.
    """
    ns, Nk, nao = obj_k.shape[:3]
    if dim == 3:
        nk = int(np.cbrt(Nk))
        # rmesh = ft.construct_symmetric_rmesh(nk, nk, nk)
        rmesh = ft.construct_rmesh(nk, nk, nk)
    elif dim == 2:
        nk = int(np.sqrt(Nk))
        # rmesh = ft.construct_symmetric_rmesh(nk, nk, 1)
        rmesh = ft.construct_rmesh(nk, nk, 1)
    else:
        raise ValueError(
            "Wannier interpolation only supports 3D and 2D systems."
        )

    fkr, frk = ft.compute_fourier_coefficients(kmesh, rmesh)
    weights = [1] * kmesh.shape[0]
    obj_i = np.array([ft.k_to_real(frk, obj_k[s], weights) for s in range(ns)])
    if debug:
        center = np.where(np.all(rmesh == (0., 0., 0.), axis=1))[0][0]
        for i in range(nk):
            print("obj_i[", i-nk//2, ", 0, 0] = ")
            print(np.diag(obj_i[0, center - nk//2+i].real))

    fkr_int, frk_int = ft.compute_fourier_coefficients(kpts_inter, rmesh)
    obj_k_int = np.array([ft.real_to_k(fkr_int, obj_i[s]) for s in range(ns)])

    if hermi:
        error = 0.0
        for s in range(ns):
            for ik in range(kpts_inter.shape[0]):
                obj = obj_k_int[s, ik]
                obj_sym = 0.5 * (obj + obj.conj().T)
                error = max(error, np.max(np.abs(obj_sym - obj)))
                obj_k_int[s, ik] = obj_sym
        print("The largest Hermitization error = ", error)

    return obj_k_int


# Only work for full_bz object
# TODO merge interpolate_tk_object and interpolate
def interpolate_tk_object(
    obj_tk, kmesh, kpts_inter, dim=3, hermi=False, debug=False
):
    """Interpolate dynamic obj_k[nts, ns, nk, nao, nao] from kmesh to
    kpts_inter using Wannier interpolation
    """
    nts, ns, Nk, nao = obj_tk.shape[:4]
    if dim == 3:
        nk = int(np.cbrt(Nk))
        # rmesh = ft.construct_symmetric_rmesh(nk, nk, nk)
        rmesh = ft.construct_rmesh(nk, nk, nk)
    elif dim == 2:
        nk = int(np.sqrt(Nk))
        # rmesh = ft.construct_symmetric_rmesh(nk, nk, 1)
        rmesh = ft.construct_rmesh(nk, nk, 1)
    else:
        raise ValueError(
            "Wannier interpolation only supports 3D and 2D systems."
        )
    fkr, frk = ft.compute_fourier_coefficients(kmesh, rmesh)
    weights = [1]*kmesh.shape[0]
    obj_ti = np.array(
        [
            ft.k_to_real(
                frk, obj_tk[it, s],
                weights
            ) for it in range(nts) for s in range(ns)
        ]
    )

    if debug:
        center = np.where(np.all(rmesh == (0., 0., 0.), axis=1))[0][0]
        for i in range(nk):
            print("obj_i[", i - nk // 2, ", 0, 0] = ")
            print(np.diag(obj_ti[0, center - nk // 2 + i].real))

    fkr_int, frk_int = ft.compute_fourier_coefficients(kpts_inter, rmesh)
    obj_tk_int = np.array(
        [ft.real_to_k(fkr_int, obj_ti[its]) for its in range(nts * ns)]
    )

    if hermi:
        error = 0.0
        for its in range(nts*ns):
            for ik in range(kpts_inter.shape[0]):
                obj = obj_tk_int[its, ik]
                obj_sym = 0.5 * (obj + obj.conj().T)
                error = max(error, np.max(np.abs(obj_sym - obj)))
                obj_tk_int[its, ik] = obj_sym
        print("The largest Hermitization error = ", error)
    obj_tk_int = obj_tk_int.reshape(nts, ns, kpts_inter.shape[0], nao, nao)

    return obj_tk_int


# Only work for full_bz object
def interpolate_G(
    Fk, Sigma_tk, mu, Sk, kmesh, kpts_inter, ir, dim=3,
    hermi=False, debug=False
):
    ns, Nk, nao = Fk.shape[:3]
    if dim != 3 and dim != 2:
        raise ValueError(
            "Wannier interpolation only supports 3D and 2D systems."
        )
    nts = ir.nts

    if Sigma_tk is not None:
        assert nts == Sigma_tk.shape[0], \
            "Number of imaginary time points mismatches."

    if Sk is not None:
        print("Interpolating overlap...")
        Sk_int = interpolate(Sk, kmesh, kpts_inter, dim, hermi, debug)
    else:
        Sk_int = None

    print("Interpolating Fock...")
    Fk_int = interpolate(Fk, kmesh, kpts_inter, dim, hermi, debug)
    # FIXME Too memory demanding and too slow as well.
    if Sigma_tk is not None:
        print("Interpolating self-energy...")
        Sigma_tk_int = interpolate_tk_object(
            Sigma_tk, kmesh, kpts_inter, dim, hermi, debug
        )
    else:
        Sigma_tk_int = None

    # Optional: Orthogonalization before Dyson

    # Solve Dyson
    Gtk_int = dyson.solve_dyson(Fk_int, Sk_int, Sigma_tk_int, mu, ir)

    return Gtk_int, Sigma_tk_int, ir.tau_mesh, Fk_int, Sk_int
